from pathlib import Path
from typing import Dict, Any

from av1an.commandtypes import Command


class Chunk:
    """
    Chunk class. Stores information relating to a chunk. The command that gets the chunk and the encoding commands
    to be run on this chunk.
    """

    def __init__(
        self,
        temp: Path,
        index: int,
        ffmpeg_gen_cmd: Command,
        output_ext: str,
        size: int,
        frames: int,
    ):
        """
        Chunk class constructor

        Note: while ffmpeg_gen_cmd is a Command, it must be serializable, so Path can't be used

        :param temp: The temp directory
        :param index: the index of this chunk
        :param ffmpeg_gen_cmd: the ffmpeg command that will generate this chunk
        :param output_ext: the output extension after encoding
        :param size: the size of this chunk. used for sorting
        :param frames: the number of frames in this chunk
        """
        self.index: int = index
        self.ffmpeg_gen_cmd: Command = ffmpeg_gen_cmd
        self.size: int = size
        self.temp: Path = temp
        self.frames: int = frames
        self.output_ext: str = output_ext
        self.per_shot_target_quality_cq = None
        self.per_frame_target_quality_q_list = None

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts this chunk to a dictionary for easy json serialization

        :return: A dictionary
        """
        return {
            "index": self.index,
            "ffmpeg_gen_cmd": self.ffmpeg_gen_cmd,
            "size": self.size,
            "frames": self.frames,
            "output_ext": self.output_ext,
            "per_shot_target_quality_cq": self.per_shot_target_quality_cq,
        }

    @property
    def fake_input_path(self) -> Path:
        """
        Returns the mkv chunk file that would have been for this chunk in the old chunk system.
        Ex: .temp/split/00000.mkv

        :return: a path
        """
        return (self.temp / "split") / f"{self.name}.mkv"

    @property
    def output_path(self) -> Path:
        """
        Gets the path of this chunk after being encoded with an extension
        Ex: Path('.temp/encode/00000.ivf')

        :return: the Path of this encoded chunk
        """
        return (self.temp / "encode") / f"{self.name}.{self.output_ext}"

    @property
    def output(self) -> str:
        """
        Gets the posix string of this chunk's output_path (with extension)
        See: Chunk.output_path
        Ex: '.temp/encode/00000.ivf'

        :return: the string of this chunk's output path
        """
        return self.output_path.as_posix()

    @property
    def fpf(self) -> str:
        """
        Gets the posix string of this chunks first pass file without an extension
        Ex: '.temp/split/00000_fpf'

        :return: the string of this chunk's first pass file (no extension)
        """
        fpf_file = (self.temp / "split") / f"{self.name}_fpf"
        return fpf_file.as_posix()

    @property
    def name(self) -> str:
        """
        Gets the name of this chunk. It is the index zero padded to length 5 as a string.
        Ex: '00000'

        :return: the name of this chunk as a string
        """
        return str(self.index).zfill(5)

    @staticmethod
    def create_from_dict(d: dict, temp):
        """
        Creates a chunk from a dictionary.
        See: Chunk.to_dict

        :param d: the dictionary
        :param temp: the temp directory
        :return: A Chunk from the dictionary
        """
        chunk = Chunk(
            temp,
            d["index"],
            d["ffmpeg_gen_cmd"],
            d["output_ext"],
            d["size"],
            d["frames"],
        )
        chunk.per_shot_target_quality_cq = d["per_shot_target_quality_cq"]
        return chunk

    def make_q_file(self, q_list):
        qfile = self.fake_input_path.with_name(f"q_file_{self.name}").with_suffix(
            ".txt"
        )
        with open(qfile, "w") as fl:
            text = ""

            for x in q_list:
                text += str(x) + "\n"

            fl.write(text)
        return qfile



import sys

from collections import deque
from typing import Iterable
from subprocess import Popen

from av1an.chunk import Chunk
from av1an.encoder import ENCODERS

from av1an.project import Project
from av1an.logger import log


def process_pipe(pipe, chunk: Chunk, utility: Iterable[Popen]):
    encoder_history = deque(maxlen=20)
    while True:
        line = pipe.stdout.readline().strip()
        if len(line) == 0 and pipe.poll() is not None:
            break
        if len(line) == 0:
            continue
        if line:
            encoder_history.append(line)

    for u_pipe in utility:
        if u_pipe.poll() is None:
            u_pipe.kill()

    if pipe.returncode != 0 and pipe.returncode != -2:
        msg1 = f"Encoder encountered an error: {pipe.returncode}"
        msg2 = f"Chunk: {chunk.index}" + "\n".join(encoder_history)
        log(msg1, msg2)
        tb = sys.exc_info()[2]
        raise RuntimeError("Error in processing encoding pipe").with_traceback(tb)


def process_encoding_pipe(
    pipe, encoder, counter, chunk: Chunk, utility: Iterable[Popen]
):
    encoder_history = deque(maxlen=20)
    frame = 0
    enc = ENCODERS[encoder]
    while True:
        line = pipe.stdout.readline().strip()

        if len(line) == 0 and pipe.poll() is not None:
            break

        if len(line) == 0:
            continue

        match = enc.match_line(line)

        if match:
            new = int(match.group(1))
            if new > frame:
                counter.update(new - frame)
                frame = new

        if line:
            encoder_history.append(line)

    for u_pipe in utility:
        if u_pipe.poll() is None:
            u_pipe.kill()

    if pipe.returncode != 0 and pipe.returncode != -2:  # -2 is Ctrl+C for aom
        msg1 = f"Encoder encountered an error: {pipe.returncode}"
        msg2 = f"Chunk: {chunk.index}"
        msg3 = "\n".join(encoder_history)
        log(msg1, msg2, msg3)
        print(f"::{msg1}\n::{msg2}\n::{msg3}")
        tb = sys.exc_info()[2]
        raise RuntimeError("Error in processing encoding pipe").with_traceback(tb)


def tqdm_bar(
    a: Project, c: Chunk, encoder, counter, frame_probe_source, passes, current_pass
):
    enc = ENCODERS[encoder]
    pipe, utility = enc.make_pipes(a, c, passes, current_pass, c.output)

    if encoder in ("aom", "vpx", "rav1e", "x265", "x264", "vvc", "svt_av1"):
        process_encoding_pipe(pipe, encoder, counter, c, utility)

    if encoder in ("svt_vp9"):
        # SVT-VP9 is special
        process_pipe(pipe, c, utility)
        counter.update(frame_probe_source // passes)



#!/bin/env python

import json
import shlex
import subprocess
import sys
from collections import deque

from pathlib import Path
from subprocess import PIPE, STDOUT

import numpy as np
from math import log10, ceil, floor
from math import log as ln

from av1an.logger import log

try:
    import matplotlib
    from matplotlib import pyplot as plt

    matplotlib.use("Agg")
except ImportError:
    matplotlib = None
    plt = None

from av1an.manager.Pipes import process_pipe
from av1an.chunk import Chunk


class VMAF:
    def __init__(self, n_threads=0, model=None, res=None, vmaf_filter=None):
        self.n_threads = f":n_threads={n_threads}" if n_threads else ""
        self.model = f":model_path={model}" if model else ""
        self.res = res if res else "1920x1080"
        self.vmaf_filter = vmaf_filter + "," if vmaf_filter else ""
        self.validate_vmaf()

    def validate_vmaf(self):
        """
        Test run of ffmpeg for validating that ffmpeg/libmaf/models properly setup
        """

        if self.model or self.n_threads:
            add = f"={self.model}{self.n_threads}"
        else:
            add = ""

        cmd = f" ffmpeg -hide_banner -filter_complex testsrc=duration=1:size=1920x1080:rate=1[B];testsrc=duration=1:size=1920x1080:rate=1[A];[B][A]libvmaf{add} -t 1  -f null - ".split()

        pipe = subprocess.Popen(
            cmd, stdout=PIPE, stderr=STDOUT, universal_newlines=True
        )

        encoder_history = deque(maxlen=30)

        while True:
            line = pipe.stdout.readline().strip()
            if len(line) == 0 and pipe.poll() is not None:
                break
            if len(line) == 0:
                continue
            if line:
                encoder_history.append(line)

        if pipe.returncode != 0 and pipe.returncode != -2:
            msg1, msg2 = f"VMAF validation error: {pipe.returncode}", "\n".join(
                encoder_history
            )
            log(msg1, msg2)
            print(f"::{msg1}\n::{msg2}")
            sys.exit()

    @staticmethod
    def read_json(file):
        """
        Reads file and return dictionary of it's contents

        :return: Vmaf file dictionary
        :rtype: dict
        """
        with open(file, "r") as f:
            fl = json.load(f)
            return fl

    def get_vmaf_motion(self):
        """
        Runs vmaf_motion filter on chunk and returns average score
        """

        cmd = [
            "ffmpeg",
            "-loglevel",
            "error",
            "-y",
            "-hide_banner",
            "-r",
            "60",
            "-i",
            "-",
            "-vf",
            "vmafmotion",
            "-f",
            "null",
            "-",
        ]

        print(cmd)

    def call_vmaf(
        self, chunk: Chunk, encoded: Path, vmaf_rate: int = None, fl_path: Path = None
    ):
        """
        Runs vmaf for Av1an
        """
        cmd = ""

        if fl_path is None:
            fl_path = chunk.fake_input_path.with_name(encoded.stem).with_suffix(".json")
        fl = fl_path.as_posix()

        cmd_in = (
            "ffmpeg",
            "-loglevel",
            "error",
            "-y",
            "-thread_queue_size",
            "1024",
            "-hide_banner",
            "-r",
            "60",
            "-i",
            encoded.as_posix(),
            "-r",
            "60",
            "-i",
            "-",
        )

        filter_complex = ("-filter_complex",)

        # Change framerate of comparison to framerate of probe
        n_subsamples = f":n_subsample={vmaf_rate}" if vmaf_rate else ""

        distorted = f"[0:v]scale={self.res}:flags=bicubic:force_original_aspect_ratio=decrease,setpts=PTS-STARTPTS[distorted];"

        ref = fr"[1:v]{self.vmaf_filter}scale={self.res}:flags=bicubic:force_original_aspect_ratio=decrease,setpts=PTS-STARTPTS[ref];"

        vmaf_filter = f"[distorted][ref]libvmaf=log_fmt='json'{n_subsamples}:eof_action=endall:log_path={shlex.quote(fl)}{self.model}{self.n_threads}"

        cmd_out = ("-f", "null", "-")

        cmd = (*cmd_in, *filter_complex, distorted + ref + vmaf_filter, *cmd_out)

        ffmpeg_gen_pipe = subprocess.Popen(
            chunk.ffmpeg_gen_cmd, stdout=PIPE, stderr=STDOUT
        )

        pipe = subprocess.Popen(
            cmd,
            stdin=ffmpeg_gen_pipe.stdout,
            stdout=PIPE,
            stderr=STDOUT,
            universal_newlines=True,
        )
        utility = (ffmpeg_gen_pipe,)
        process_pipe(pipe, chunk, utility)

        return fl_path

    @staticmethod
    def get_percentile(scores, percent):
        """
        Find the percentile of a list of values.
        :param scores: - is a list of values. Note N MUST BE already sorted.
        :param percent: - a float value from 0.0 to 1.0.
        :return: - the percentile of the values
        """
        scores = sorted(scores)
        key = lambda x: x

        k = (len(scores) - 1) * percent
        f = floor(k)
        c = ceil(k)
        if f == c:
            return key(scores[int(k)])

        d0 = (scores[int(f)]) * (c - k)
        d1 = (scores[int(c)]) * (k - f)
        return d0 + d1

    @staticmethod
    def transform_vmaf(vmaf):
        if vmaf < 99.99:
            return -ln(1 - vmaf / 100)
        else:
            # return -ln(1-99.99/100)
            return 9.210340371976184

    @staticmethod
    def read_vmaf_with_motion_compensation(file, percentile=0):
        """Reads vmaf file with vmaf scores in it and return N percentile score from it.

        :return: N percentile score
        :rtype: float
        """

        jsn = VMAF.read_json(file)

        vmafs = sorted([x["metrics"]["vmaf"] for x in jsn["frames"]])
        motion = np.average([x["metrics"]["motion2"] for x in jsn["frames"]])
        print(round(motion, 1))
        percentile = percentile if percentile != 0 else 0.25
        score = VMAF.get_percentile(vmafs, percentile)

        return round(score, 2)

    @staticmethod
    def read_weighted_vmaf(file, percentile=0):
        """Reads vmaf file with vmaf scores in it and return N percentile score from it.

        :return: N percentile score
        :rtype: float
        """

        jsn = VMAF.read_json(file)

        vmafs = sorted([x["metrics"]["vmaf"] for x in jsn["frames"]])

        percentile = percentile if percentile != 0 else 0.25
        score = VMAF.get_percentile(vmafs, percentile)

        return round(score, 2)

    def get_vmaf_file(self, source: Path, encoded: Path):
        """
        Running vmaf on 2 files and returning file
        """

        if not all((isinstance(source, Path), isinstance(encoded, Path))):
            source = Path(source)
            encoded = Path(encoded)

        fl_path = encoded.with_name(f"{encoded.stem}_vmaflog").with_suffix(".json")

        # call_vmaf takes a chunk, so make a chunk of the entire source
        ffmpeg_gen_cmd = [
            "ffmpeg",
            "-y",
            "-hide_banner",
            "-loglevel",
            "error",
            "-i",
            source.as_posix(),
            "-f",
            "yuv4mpegpipe",
            "-",
        ]

        input_chunk = Chunk("", 0, ffmpeg_gen_cmd, "", 0, 0)

        scores = self.call_vmaf(input_chunk, encoded, 0, fl_path=fl_path)
        return scores

    def get_vmaf_json(self, source: Path, encoded: Path):
        """
        Returning dictionary from vmaf json
        """
        fl = self.get_vmaf_file(source, encoded)
        js = self.read_json(fl)
        return js

    def get_vmaf_score(self, source: Path, encoded: Path, percentile=50):
        """
        Returning calculated vmaf score
        Posible to set percentile, default 50

        :rtype: float
        """
        js = self.get_vmaf_json(source, encoded)
        score = np.average([x["metrics"]["vmaf"] for x in js["frames"]])
        return score

    def plot_vmaf(self, source: Path, encoded: Path, args):
        """
        Making VMAF plot after encode is done
        """

        print(":: VMAF Run..", end="\r")

        fl_path = encoded.with_name(f"{encoded.stem}_vmaflog").with_suffix(".json")

        # call_vmaf takes a chunk, so make a chunk of the entire source
        ffmpeg_gen_cmd = [
            "ffmpeg",
            "-y",
            "-hide_banner",
            "-loglevel",
            "error",
            "-i",
            source.as_posix(),
            *args.pix_format,
            "-f",
            "yuv4mpegpipe",
            "-",
        ]

        input_chunk = Chunk(args.temp, 0, ffmpeg_gen_cmd, "", 0, 0)

        scores = self.call_vmaf(input_chunk, encoded, 0, fl_path=fl_path)

        if not scores.exists():

            print(f"Vmaf calculation failed for chunks:\n {source.name} {encoded.stem}")
            sys.exit()

        file_path = encoded.with_name(f"{encoded.stem}_plot").with_suffix(".png")
        self.plot_vmaf_score_file(scores, file_path)

    def plot_vmaf_score_file(self, scores: Path, plot_path: Path):
        """
        Read vmaf json and plot VMAF values for each frame
        """
        if plt is None:
            log(
                f"Matplotlib is not installed or could not be loaded, aborting plot_vmaf"
            )
            return

        perc_1 = self.read_weighted_vmaf(scores, 0.01)
        perc_25 = self.read_weighted_vmaf(scores, 0.25)
        perc_75 = self.read_weighted_vmaf(scores, 0.75)
        mean = self.read_weighted_vmaf(scores, 0.50)

        with open(scores) as f:
            file = json.load(f)
            vmafs = [x["metrics"]["vmaf"] for x in file["frames"]]
            plot_size = len(vmafs)

        figure_width = 3 + round((4 * log10(plot_size)))
        plt.figure(figsize=(figure_width, 5))

        plt.plot([1, plot_size], [perc_1, perc_1], "-", color="red")
        plt.annotate(f"1%: {perc_1}", xy=(0, perc_1), color="red")

        plt.plot([1, plot_size], [perc_25, perc_25], ":", color="orange")
        plt.annotate(f"25%: {perc_25}", xy=(0, perc_25), color="orange")

        plt.plot([1, plot_size], [perc_75, perc_75], ":", color="green")
        plt.annotate(f"75%: {perc_75}", xy=(0, perc_75), color="green")

        plt.plot([1, plot_size], [mean, mean], ":", color="black")
        plt.annotate(f"Mean: {mean}", xy=(0, mean), color="black")

        for i in range(0, 100):
            plt.axhline(i, color="grey", linewidth=0.4)
            if i % 5 == 0:
                plt.axhline(i, color="black", linewidth=0.6)

        plt.plot(
            range(plot_size),
            vmafs,
            label=f"Frames: {plot_size}\nMean:{mean}\n"
            f"1%: {perc_1} \n25%: {perc_25} \n75%: {perc_75}",
            linewidth=0.7,
        )
        plt.ylabel("VMAF")
        plt.legend(
            loc="lower right",
            markerscale=0,
            handlelength=0,
            fancybox=True,
        )
        plt.ylim(int(perc_1), 100)
        plt.tight_layout()
        plt.margins(0)

        # Save
        plt.savefig(plot_path, dpi=250)



import re
from subprocess import PIPE
from pathlib import Path
from shlex import split
from subprocess import run, Popen

VS_EXTENSIONS = [".vpy", ".py"]


def is_vapoursynth(path: Path):
    return path.suffix in VS_EXTENSIONS


def frame_probe_vspipe(source: Path):
    """
    Get frame count from vspipe.
    :param: source: Path to input vapoursynth (vpy/py) file
    """
    cmd = ["vspipe", "-i", source.as_posix(), "-"]
    r = run(cmd, capture_output=True)
    matches = re.findall(
        r"Frames:\s*([0-9]+)\s", r.stderr.decode("utf-8") + r.stdout.decode("utf-8")
    )
    frames = int(matches[-1])
    return frames


def create_vs_file(temp: Path, source, chunk_method):
    """
    Creates vs pipe file or returns file if it exists
    """

    load_script = temp / "split" / "loadscript.vpy"

    if load_script.exists():
        return load_script

    if chunk_method == "vs_ffms2":
        cache_file = (temp / "split" / "cache.ffindex").resolve()
        script = (
            "from vapoursynth import core\n"
            'core.ffms2.Source(r"{}", cachefile="{}").set_output()'
        )
    else:
        cache_file = (temp / "split" / "cache.lwi").resolve().as_posix()
        script = (
            "from vapoursynth import core\n"
            'core.lsmas.LWLibavSource(r"{}", cachefile="{}").set_output()'
        )

    with open(load_script, "w+", encoding='utf-8') as file:
        file.write(script.format(Path(source).resolve(), cache_file))

    cache_generation = f"vspipe -i {load_script.as_posix()} -i -"
    d = Popen(split(cache_generation), stdout=PIPE, stderr=PIPE).wait()

    return load_script


def compose_vapoursynth_pipe(source: Path, fifo: Path = None):
    return ["vspipe", "-y", source.as_posix(), fifo.as_posix() if fifo else "-"]



import os
import platform
import shlex
import subprocess
import sys
from pathlib import Path
from subprocess import PIPE, STDOUT

from av1an.logger import log

if platform.system() == "Linux":
    import resource


def vvc_concat(temp: Path, output: Path):
    """
    Concatenates vvc files

    :param temp: the temp directory
    :param output: the output video
    :return: None
    """
    encode_files = sorted((temp / "encode").iterdir())
    bitstreams = [x.as_posix() for x in encode_files]
    bitstreams = " ".join(bitstreams)
    cmd = f"vvc_concat  {bitstreams} {output.as_posix()}"

    subprocess.run(cmd, shell=True, check=True)


def concatenate_ffmpeg(temp: Path, output: Path, encoder: str):
    """
    Uses ffmpeg to concatenate encoded segments into the final file

    :param temp: the temp directory
    :param output: the final output file
    :param encoder: the encoder
    :return: None
    """

    log("Concatenating")

    with open(temp / "concat", "w") as f:

        encode_files = sorted((temp / "encode").iterdir())
        f.writelines(
            f'file {shlex.quote("file:"+str(file.absolute()))}\n'
            for file in encode_files
        )

    # Add the audio/subtitles/else file if one was extracted from the input
    audio_file = temp / "audio.mkv"
    if audio_file.exists():
        audio = ("-i", audio_file.as_posix(), "-c", "copy", "-map", "1")
    else:
        audio = ()

    if encoder == "x265":

        cmd = [
            "ffmpeg",
            "-y",
            "-fflags",
            "+genpts",
            "-hide_banner",
            "-loglevel",
            "error",
            "-f",
            "concat",
            "-safe",
            "0",
            "-i",
            (temp / "concat").as_posix(),
            *audio,
            "-c",
            "copy",
            "-movflags",
            "frag_keyframe+empty_moov",
            "-map",
            "0",
            "-f",
            "mp4",
            output.as_posix(),
        ]
        concat = subprocess.run(cmd, stdout=PIPE, stderr=STDOUT, check=True).stdout

    else:
        cmd = [
            "ffmpeg",
            "-y",
            "-hide_banner",
            "-loglevel",
            "error",
            "-f",
            "concat",
            "-safe",
            "0",
            "-i",
            (temp / "concat").as_posix(),
            *audio,
            "-c",
            "copy",
            "-map",
            "0",
            output.as_posix(),
        ]

        concat = subprocess.run(cmd, stdout=PIPE, stderr=STDOUT, check=True).stdout

    if len(concat) > 0:
        log(concat.decode())
        print(concat.decode())
        tb = sys.exc_info()[2]
        raise RuntimeError.with_traceback(tb)


def concatenate_mkvmerge(temp: Path, output):
    """
    Uses mkvmerge to concatenate encoded segments into the final file

    :param temp: the temp directory
    :param output: the final output file
    :return: None
    """

    log("Concatenating")

    output = shlex.quote(output.as_posix())

    encode_files = sorted(
        (temp / "encode").iterdir(),
        key=lambda x: int(x.stem) if x.stem.isdigit() else x.stem,
    )
    encode_files = [shlex.quote(f.as_posix()) for f in encode_files]

    if platform.system() == "Linux":
        file_limit, _ = resource.getrlimit(resource.RLIMIT_NOFILE)
        cmd_limit = os.sysconf(os.sysconf_names["SC_ARG_MAX"])
    else:
        file_limit = -1
        cmd_limit = 32767

    audio_file = temp / "audio.mkv"
    audio = audio_file.as_posix() if audio_file.exists() else ""

    if len(encode_files) > 1:
        encode_files = [
            _concatenate_mkvmerge(encode_files, output, file_limit, cmd_limit)
        ]

    cmd = ["mkvmerge", "-o", output, encode_files[0]]

    if audio:
        cmd.append(audio)

    concat = subprocess.Popen(cmd, stdout=PIPE, universal_newlines=True)
    message, _ = concat.communicate()
    concat.wait()

    if concat.returncode != 0:
        log(message)
        print(message)
        tb = sys.exc_info()[2]
        raise RuntimeError.with_traceback(tb)

    # remove temporary files used by recursive concat
    if os.path.exists("{}.tmp0.mkv".format(output)):
        os.remove("{}.tmp0.mkv".format(output))

    if os.path.exists("{}.tmp1.mkv".format(output)):
        os.remove("{}.tmp1.mkv".format(output))


def _concatenate_mkvmerge(files, output, file_limit, cmd_limit, flip=False):
    tmp_out = "{}.tmp{}.mkv".format(output, int(flip))
    cmd = ["mkvmerge", "-o", tmp_out, files[0]]

    remaining = []
    for i, file in enumerate(files[1:]):
        new_cmd = cmd + ["+{}".format(file)]
        if sum(len(s) for s in new_cmd) < cmd_limit and (
            file_limit == -1 or i < max(1, file_limit - 10)
        ):
            cmd = new_cmd
        else:
            remaining = files[i + 1 :]
            break

    concat = subprocess.Popen(cmd, stdout=PIPE, universal_newlines=True)
    message, _ = concat.communicate()
    concat.wait()

    if concat.returncode != 0:
        log(message)
        print(message)
        tb = sys.exc_info()[2]
        raise RuntimeError.with_traceback(tb)

    if len(remaining) > 0:
        return _concatenate_mkvmerge(
            [tmp_out] + remaining, output, file_limit, cmd_limit, not flip
        )
    return tmp_out



#!/bin/env python
import numpy as np


def process_inputs(inputs):
    # Check input file for being valid
    if not inputs:
        print("No input file")
        exit()

    input_list = []

    for item in inputs:
        if item.is_dir():
            new_inputs = [
                x
                for x in item.iterdir()
                if x.suffix in (".mkv", ".mp4", ".mov", ".avi", ".flv", ".m2ts")
            ]
            input_list.extend(new_inputs)
        else:
            input_list.append(item)

    valid = np.array([i.exists() for i in input_list])

    if not all(valid):
        print(
            f'File(s) do not exist: {", ".join([str(input_list[i]) for i in np.where(not valid)[0]])}'
        )
        exit()

    return input_list



import json
from pathlib import Path
from typing import List
from subprocess import Popen, DEVNULL
from shlex import split

from av1an.project import Project
from av1an.chunk import Chunk
from av1an.encoder import ENCODERS
from av1an.ffmpeg import get_keyframes
from av1an.logger import log
from av1an.resume import read_done_data
from av1an.split import segment
from av1an.utils import terminate
from av1an.vapoursynth import create_vs_file


def save_chunk_queue(temp: Path, chunk_queue: List[Chunk]) -> None:
    """
    Writes the chunk queue to the chunks.json file

    :param temp: the temp directory
    :param chunk_queue: the chunk queue
    :return: None
    """
    chunk_dicts = [c.to_dict() for c in chunk_queue]
    with open(temp / "chunks.json", "w") as file:
        json.dump(chunk_dicts, file)


def read_chunk_queue(temp: Path) -> List[Chunk]:
    """
    Reads the chunk queue from the chunks.json file

    :param temp: the temp directory
    :return: the chunk queue
    """
    with open(temp / "chunks.json", "r") as file:
        chunk_dicts = json.load(file)
    return [Chunk.create_from_dict(cd, temp) for cd in chunk_dicts]


def load_or_gen_chunk_queue(
    project: Project, resuming: bool, split_locations: List[int]
) -> List[Chunk]:
    """
    If resuming, loads the chunk queue and removes already done chunks or
    creates a chunk queue and saves it for resuming later.

    :param project: the Project
    :param resuming: if we are resuming
    :param split_locations: a list of frames to split on
    :return: A chunk queue (list of chunks)
    """
    # if resuming, read chunks from file and remove those already done
    if resuming:
        chunk_queue = read_chunk_queue(project.temp)
        done_chunk_names = read_done_data(project.temp)["done"].keys()
        chunk_queue = [c for c in chunk_queue if c.name not in done_chunk_names]
        return chunk_queue

    # create and save
    chunk_queue = create_encoding_queue(project, split_locations)
    save_chunk_queue(project.temp, chunk_queue)

    return chunk_queue


def create_encoding_queue(project: Project, split_locations: List[int]) -> List[Chunk]:
    """
    Creates a list of chunks using the cli option chunk_method specified

    :param project: Project
    :param split_locations: a list of frames to split on
    :return: A list of chunks
    """
    chunk_method_gen = {
        "segment": create_video_queue_segment,
        "select": create_video_queue_select,
        "vs_ffms2": create_video_queue_vsffms2,
        "vs_lsmash": create_video_queue_vslsmash,
        "hybrid": create_video_queue_hybrid,
    }
    chunk_queue = chunk_method_gen[project.chunk_method](project, split_locations)

    # Sort largest first so chunks that take a long time to encode start first
    chunk_queue.sort(key=lambda c: c.size, reverse=True)
    return chunk_queue


def create_video_queue_hybrid(
    project: Project, split_locations: List[int]
) -> List[Chunk]:
    """
    Create list of chunks using hybrid segment-select approach

    :param project: the Project
    :param split_locations: a list of frames to split on
    :return: A list of chunks
    """
    keyframes = get_keyframes(project.input)

    end = [project.get_frames()]

    splits = [0] + split_locations + end

    segments_list = list(zip(splits, splits[1:]))
    to_split = [x for x in keyframes if x in splits]
    segments = []

    # Make segments
    segment(project.input, project.temp, to_split[1:])
    source_path = project.temp / "split"
    queue_files = [x for x in source_path.iterdir() if x.suffix == ".mkv"]
    queue_files.sort(key=lambda p: p.stem)

    kf_list = list(zip(to_split, to_split[1:] + end))
    for f, (x, y) in zip(queue_files, kf_list):
        to_add = [
            (f, [s[0] - x, s[1] - x])
            for s in segments_list
            if s[0] >= x and s[1] <= y and s[0] - x < s[1] - x
        ]
        segments.extend(to_add)

    chunk_queue = [
        create_select_chunk(project, index, file, *cb)
        for index, (file, cb) in enumerate(segments)
    ]
    return chunk_queue


def create_video_queue_vsffms2(
    project: Project, split_locations: List[int]
) -> List[Chunk]:
    return create_video_queue_vs(project, split_locations)


def create_video_queue_vslsmash(
    project: Project, split_locations: List[int]
) -> List[Chunk]:
    return create_video_queue_vs(project, split_locations)


def create_video_queue_vs(project: Project, split_locations: List[int]) -> List[Chunk]:
    """
    Create a list of chunks using vspipe and ffms2 for frame accurate seeking

    :param project: the Project
    :param split_locations: a list of frames to split on
    :param script: source filter script to use with vspipe (ignored with vs input)
    :return: A list of chunks
    """
    # add first frame and last frame
    last_frame = project.get_frames()
    split_locs_fl = [0] + split_locations + [last_frame]

    # pair up adjacent members of this list ex: [0, 10, 20, 30] -> [(0, 10), (10, 20), (20, 30)]
    chunk_boundaries = zip(split_locs_fl, split_locs_fl[1:])

    source_file = project.input.absolute().as_posix()
    if project.is_vs:
        vs_script = project.input
    else:
        vs_script = create_vs_file(project.temp, source_file, project.chunk_method)

    chunk_queue = [
        create_vs_chunk(project, index, vs_script, *cb)
        for index, cb in enumerate(chunk_boundaries)
    ]

    return chunk_queue


def create_vs_chunk(
    project: Project, index: int, vs_script: Path, frame_start: int, frame_end: int
) -> Chunk:
    """
    Creates a chunk using vspipe

    :param project: the Project
    :param load_script: the path to the .vpy script for vspipe
    :param index: the index of the chunk
    :param frame_start: frame that this chunk should start on (0-based, inclusive)
    :param frame_end: frame that this chunk should end on (0-based, exclusive)
    :return: a Chunk
    """
    assert frame_end > frame_start, "Can't make a chunk with <= 0 frames!"

    frames = frame_end - frame_start
    frame_end -= 1  # the frame end boundary is actually a frame that should be included in the next chunk

    vspipe_gen_cmd = [
        "vspipe",
        vs_script.resolve().as_posix(),
        "-y",
        "-",
        "-s",
        str(frame_start),
        "-e",
        str(frame_end),
    ]
    extension = ENCODERS[project.encoder].output_extension
    size = frames  # use the number of frames to prioritize which chunks encode first, since we don't have file size

    chunk = Chunk(project.temp, index, vspipe_gen_cmd, extension, size, frames)

    return chunk


def create_video_queue_select(
    project: Project, split_locations: List[int]
) -> List[Chunk]:
    """
    Create a list of chunks using the select filter

    :param project: the Project
    :param split_locations: a list of frames to split on
    :return: A list of chunks
    """
    # add first frame and last frame
    last_frame = project.get_frames()
    split_locs_fl = [0] + split_locations + [last_frame]

    # pair up adjacent members of this list ex: [0, 10, 20, 30] -> [(0, 10), (10, 20), (20, 30)]
    chunk_boundaries = zip(split_locs_fl, split_locs_fl[1:])

    chunk_queue = [
        create_select_chunk(project, index, project.input, *cb)
        for index, cb in enumerate(chunk_boundaries)
    ]

    return chunk_queue


def create_select_chunk(
    project: Project, index: int, src_path: Path, frame_start: int, frame_end: int
) -> Chunk:
    """
    Creates a chunk using ffmpeg's select filter

    :param project: the Project
    :param src_path: the path of the entire unchunked source file
    :param index: the index of the chunk
    :param frame_start: frame that this chunk should start on (0-based, inclusive)
    :param frame_end: frame that this chunk should end on (0-based, exclusive)
    :return: a Chunk
    """
    assert frame_end > frame_start, "Can't make a chunk with <= 0 frames!"

    frames = frame_end - frame_start
    frame_end -= 1  # the frame end boundary is actually a frame that should be included in the next chunk

    ffmpeg_gen_cmd = [
        "ffmpeg",
        "-y",
        "-hide_banner",
        "-loglevel",
        "error",
        "-i",
        src_path.as_posix(),
        "-vf",
        f"select=between(n\\,{frame_start}\\,{frame_end}),setpts=PTS-STARTPTS",
        *project.pix_format,
        "-color_range",
        "0",
        "-f",
        "yuv4mpegpipe",
        "-",
    ]
    extension = ENCODERS[project.encoder].output_extension
    size = frames  # use the number of frames to prioritize which chunks encode first, since we don't have file size

    chunk = Chunk(project.temp, index, ffmpeg_gen_cmd, extension, size, frames)

    return chunk


def create_video_queue_segment(
    project: Project, split_locations: List[int]
) -> List[Chunk]:
    """
    Create a list of chunks using segmented files

    :param project: Project
    :param split_locations: a list of frames to split on
    :return: A list of chunks
    """

    # segment into separate files
    segment(project.input, project.temp, split_locations)

    # get the names of all the split files
    source_path = project.temp / "split"
    queue_files = [x for x in source_path.iterdir() if x.suffix == ".mkv"]
    queue_files.sort(key=lambda p: p.stem)

    if len(queue_files) == 0:
        er = "Error: No files found in temp/split, probably splitting not working"
        print(er)
        log(er)
        terminate()

    chunk_queue = [
        create_chunk_from_segment(project, index, file)
        for index, file in enumerate(queue_files)
    ]

    return chunk_queue


def create_chunk_from_segment(project: Project, index: int, file: Path) -> Chunk:
    """
    Creates a Chunk object from a segment file generated by ffmpeg

    :param project: the Project
    :param index: the index of the chunk
    :param file: the segmented file
    :return: A Chunk
    """
    ffmpeg_gen_cmd = [
        "ffmpeg",
        "-y",
        "-hide_banner",
        "-loglevel",
        "error",
        "-i",
        file.as_posix(),
        *project.pix_format,
        "-color_range",
        "0",
        "-f",
        "yuv4mpegpipe",
        "-",
    ]
    file_size = file.stat().st_size
    frames = project.get_frames()
    extension = ENCODERS[project.encoder].output_extension

    chunk = Chunk(project.temp, index, ffmpeg_gen_cmd, extension, file_size, frames)

    return chunk



#!/bin/env python

import os
import subprocess
import json
from pathlib import Path
from subprocess import PIPE, STDOUT
from typing import List, Tuple
from numpy import linspace

from .project import Project
from .scenedetection import aom_keyframes, AOM_KEYFRAMES_DEFAULT_PARAMS, pyscene, ffmpeg
from .logger import log
from .utils import terminate

# TODO: organize to single segmenting/splitting module


def split_routine(project: Project, resuming: bool) -> List[int]:
    """
    Performs the split routine. Runs pyscenedetect/aom keyframes and adds in extra splits if needed

    :param project: the Project
    :param resuming: if the encode is being resumed
    :return: A list of frames to split on
    """
    scene_file = project.temp / "scenes.txt"

    # if resuming, we already have the split file, so just read that and return
    if resuming:
        scenes, frames = read_scenes_from_file(scene_file)
        project.set_frames(frames)
        return scenes

    # Run scenedetection or skip
    if project.split_method == "none":
        log("Skipping scene detection")
        scenes = []

    # Read saved scenes:
    if project.scenes and Path(project.scenes).exists():
        log("Using Saved Scenes")
        scenes, frames = read_scenes_from_file(Path(project.scenes))
        project.set_frames(frames)

    else:
        # determines split frames with pyscenedetect or aom keyframes
        scenes = calc_split_locations(project)
        if project.scenes and Path(project.scenes).exists():
            write_scenes_to_file(scenes, project.get_frames(), Path(project.scenes))

    # Write internal scenes
    write_scenes_to_file(scenes, project.get_frames(), scene_file)

    # Applying extra splits
    if project.extra_split:
        scenes = extra_splits(project, scenes)

    # write scenes for resuming later if needed
    return scenes


def write_scenes_to_file(scenes: List[int], frames: int, scene_path: Path):
    """
    Writes a list of scenes to the a file

    :param scenes: the scenes to write
    :param scene_path: the file to write to
    :return: None
    """
    with open(scene_path, "w") as scene_file:
        data = {"scenes": scenes, "frames": frames}
        json.dump(data, scene_file)


def read_scenes_from_file(scene_path: Path) -> Tuple[int]:
    """
    Reads a list of split locations from a file

    :param scene_path: the file to read from
    :return: a list of frames to split on
    """
    with open(scene_path, "r") as scene_file:
        data = json.load(scene_file)
        return data["scenes"], data["frames"]


def segment(video: Path, temp: Path, frames: List[int]):
    """
    Uses ffmpeg to segment the video into separate files.
    Splits the video by frame numbers or copies the video if no splits are needed

    :param video: the source video
    :param temp: the temp directory
    :param frames: the split locations
    :return: None
    """

    log("Split Video")
    cmd = [
        "ffmpeg",
        "-hide_banner",
        "-y",
        "-i",
        video.absolute().as_posix(),
        "-map",
        "0:v:0",
        "-an",
        "-c",
        "copy",
        "-avoid_negative_ts",
        "1",
        "-vsync",
        "0",
    ]

    if len(frames) > 0:
        cmd.extend(
            ["-f", "segment", "-segment_frames", ",".join([str(x) for x in frames])]
        )
        cmd.append(os.path.join(temp, "split", "%05d.mkv"))
    else:
        cmd.append(os.path.join(temp, "split", "0.mkv"))
    pipe = subprocess.Popen(cmd, stdout=PIPE, stderr=STDOUT)
    while True:
        line = pipe.stdout.readline().strip()
        if len(line) == 0 and pipe.poll() is not None:
            break

    log("Split Done")


def extra_splits(project: Project, split_locations: list):
    log("Applying extra splits")

    split_locs_with_start = split_locations[:]
    split_locs_with_start.insert(0, 0)

    split_locs_with_end = split_locations[:]
    split_locs_with_end.append(project.get_frames())

    splits = list(zip(split_locs_with_start, split_locs_with_end))
    for i in splits:
        distance = i[1] - i[0]
        if distance > project.extra_split:
            to_add = distance // project.extra_split
            new_scenes = list(
                linspace(i[0], i[1], to_add + 1, dtype=int, endpoint=False)[1:]
            )
            split_locations.extend(new_scenes)

    result = [int(x) for x in sorted(split_locations)]
    log(f"Split distance: {project.extra_split}")
    log(f"New splits:{len(result)}")
    return result


def calc_split_locations(project: Project) -> List[int]:
    """
    Determines a list of frame numbers to split on with pyscenedetect or aom keyframes

    :param project: the Project
    :return: A list of frame numbers
    """
    # inherit video params from aom encode unless we are using a different encoder, then use defaults
    aom_keyframes_params = (
        project.video_params
        if (project.encoder == "aom")
        else AOM_KEYFRAMES_DEFAULT_PARAMS
    )

    sc = []

    # Splitting using PySceneDetect
    if project.split_method == "pyscene":
        log(
            f"Starting scene detection Threshold: {project.threshold}, Min_scene_length: {project.min_scene_len}"
        )
        try:
            sc = pyscene(
                project.input,
                project.threshold,
                project.min_scene_len,
                project.is_vs,
                project.temp,
                project.quiet,
            )
        except Exception as e:
            log(f"Error in PySceneDetect: {e}")
            print(f"Error in PySceneDetect{e}")
            terminate()

    # Splitting based on aom keyframe placement
    elif project.split_method == "aom_keyframes":
        stat_file = project.temp / "keyframes.log"
        sc = aom_keyframes(
            project.input,
            stat_file,
            project.min_scene_len,
            project.ffmpeg_pipe,
            aom_keyframes_params,
            project.is_vs,
            project.quiet,
        )

    elif project.split_method == "ffmpeg":
        sc = ffmpeg(
            project.input,
            project.threshold,
            project.min_scene_len,
            project.get_frames(),
            project.is_vs,
            project.temp,
        )

    # Write scenes to file
    if project.scenes:
        write_scenes_to_file(sc, project.get_frames(), project.scenes)

    return sc



#!/bin/env python

import re
import sys
from typing import List
from pathlib import Path
import cv2
import numpy as np
import hashlib

from av1an.ffmpeg import frame_probe_ffmpeg
from av1an.vapoursynth import frame_probe_vspipe, is_vapoursynth


def terminate():
    sys.exit(1)


def hash_path(s: str) -> int:
    """
    Return hash of full path to file
    :param s: string
    """
    assert isinstance(s, str)

    return str(hashlib.sha3_512(s.encode()).hexdigest())[-8:]


def get_cq(command):
    """
    Return cq values from command
    :param command: string with commands for encoder
    :return: list with frame numbers of keyframes

    """
    matches = re.findall(r"--cq-level= *([^ ]+?) ", command)
    return int(matches[-1])


def list_index_of_regex(lst: List[str], regex_str: str) -> int:
    """
    Gets the first index of the list where regex_str matches

    :param lst: the list
    :param regex_str: the regex as a string
    :return: the index where regex_str appears in the list
    :raises ValueError: if regex_str is not found
    """
    reg = re.compile(regex_str)
    for i, elem in enumerate(lst):
        if reg.match(elem):
            return i
    raise ValueError(f"{reg} is not in list")


def frame_probe_fast(source: Path, is_vs: bool = False):
    """
    Consolidated function to retrieve the number of frames from the input quickly,
    falls back on a slower (but accurate) frame count if a quick count cannot be found.

    Handles vapoursynth input as well.
    """
    total = 0
    if not is_vs:
        try:
            import vapoursynth
            from vapoursynth import core

            plugins = vapoursynth.get_core().get_plugins()
            if "systems.innocent.lsmas" in plugins:
                total = core.lsmas.LWLibavSource(
                    source.as_posix(), cache=False
                ).num_frames
                return total
        except:
            video = cv2.VideoCapture(source.as_posix())
            total = int(video.get(cv2.CAP_PROP_FRAME_COUNT))
            video.release()
    if is_vs or total < 1:
        total = frame_probe(source)

    return total


def frame_probe(source: Path):
    """
    Determines the total number of frames in a given input.

    Differentiates between a Vapoursynth script and standard video
    and delegates to vspipe or ffmpeg respectively.
    """
    if is_vapoursynth(source):
        return frame_probe_vspipe(source)

    return frame_probe_ffmpeg(source)



#!/bin/env python

import sys
import time
from pathlib import Path

# Todo: Add self testing on startup
class Logger:
    def __init__(self):
        self.set_file = False
        self.buffer = ""

    def set_path(self, file):
        self.set_file = Path(file)

    def log(self, *info):
        """Default logging function, write to file."""
        for i in info:
            if self.set_file and self.buffer:
                with open(self.set_file, "a") as logf:
                    logf.write(self.buffer)
                    self.buffer = None

            if self.set_file:
                with open(self.set_file, "a") as logf:
                    logf.write(f'[{time.strftime("%X")}] {i}\n')
            else:
                self.buffer += f'[{time.strftime("%X")}] {i}\n'


# Creating logger
logger = Logger()
log_file = logger.set_path
log = logger.log


def set_log(log_path: Path, temp):
    """Setting logging file location"""

    if log_path:
        log_path = Path(log_path)
        if log_path.suffix == "":
            log_path = log_path.with_suffix(".log")
        log_file(log_path)

    else:
        log_file(temp / "log.log")

    log(f"Av1an Started")
    log(f"Command:")
    log(f"{' '.join(sys.argv)}")



from .aom import Aom
from .rav1e import Rav1e
from .svtav1 import SvtAv1
from .svtvp9 import SvtVp9
from .vpx import Vpx
from .vvc import Vvc
from .x264 import X264
from .x265 import X265

ENCODERS = {
    'aom': Aom(),
    'rav1e': Rav1e(),
    'svt_av1': SvtAv1(),
    'svt_vp9': SvtVp9(),
    'vpx': Vpx(),
    'vvc': Vvc(),
    'x264': X264(),
    'x265': X265(),
}



import os
import re

from av1an.project import Project
from av1an.chunk import Chunk
from av1an.commandtypes import MPCommands, CommandPair, Command
from av1an.encoder.encoder import Encoder
from av1an.utils import list_index_of_regex, terminate


class Rav1e(Encoder):
    def __init__(self):
        super().__init__(
            encoder_bin="rav1e",
            encoder_help="rav1e --fullhelp",
            default_args=[
                "--tiles",
                "8",
                "--speed",
                "6",
                "--quantizer",
                "100",
                "--no-scene-detection",
            ],
            default_passes=1,
            default_q_range=(50, 140),
            output_extension="ivf",
        )

    def compose_1_pass(self, a: Project, c: Chunk, output: str) -> MPCommands:
        return [
            CommandPair(
                Encoder.compose_ffmpeg_pipe(a),
                ["rav1e", "-", "-y", *a.video_params, "--output", output],
            )
        ]

    def compose_2_pass(self, a: Project, c: Chunk, output: str) -> MPCommands:
        return [
            CommandPair(
                Encoder.compose_ffmpeg_pipe(a),
                [
                    "rav1e",
                    "-",
                    "-q",
                    "-y",
                    "--first-pass",
                    f"{c.fpf}.stat",
                    *a.video_params,
                    "--output",
                    os.devnull,
                ],
            ),
            CommandPair(
                Encoder.compose_ffmpeg_pipe(a),
                [
                    "rav1e",
                    "-",
                    "-y",
                    "--second-pass",
                    f"{c.fpf}.stat",
                    *a.video_params,
                    "--output",
                    output,
                ],
            ),
        ]

    def man_q(self, command: Command, q: int) -> Command:
        """Return command with new cq value

        :param command: old command
        :param q: new cq value
        :return: command with new cq value"""

        adjusted_command = command.copy()

        i = list_index_of_regex(adjusted_command, r"--quantizer")
        adjusted_command[i + 1] = f"{q}"

        return adjusted_command

    def match_line(self, line: str):
        """Extract number of encoded frames from line.

        :param line: one line of text output from the encoder
        :return: match object from re.search matching the number of encoded frames"""

        if "error" in line.lower():
            print("\n\nERROR IN ENCODING PROCESS\n\n", line)
            terminate()
        return re.search(r"encoded.*? ([^ ]+?) ", line)



import subprocess
import os

from math import isnan
import numpy as np
import re
from scipy import interpolate

from av1an.vmaf import VMAF
from av1an.logger import log
from av1an.commandtypes import CommandPair, Command
from av1an.chunk import Chunk
from av1an.manager.Pipes import process_pipe

try:
    import matplotlib
    from matplotlib import pyplot as plt
except ImportError:
    matplotlib = None
    plt = None


class TargetQuality:
    def __init__(self, project):
        self.vmaf_runner = VMAF(
            n_threads=project.n_threads,
            model=project.vmaf_path,
            res=project.vmaf_res,
            vmaf_filter=project.vmaf_filter,
        )
        self.n_threads = project.n_threads
        self.probing_rate = project.probing_rate
        self.probes = project.probes
        self.target = project.target_quality
        self.min_q = project.min_q
        self.max_q = project.max_q
        self.make_plots = project.vmaf_plots
        self.encoder = project.encoder
        self.ffmpeg_pipe = project.ffmpeg_pipe
        self.temp = project.temp
        self.workers = project.workers

    def per_frame_target_quality_routine(self, chunk: Chunk):
        """
        Applies per_shot_target_quality to this chunk. Determines what the cq value should be and sets the
        per_shot_target_quality_cq for this chunk

        :param project: the Project
        :param chunk: the Chunk
        :return: None
        """
        chunk.per_frame_target_quality_q_list = self.per_frame_target_quality(chunk)

    def log_probes(self, vmaf_cq, frames, name, target_q, target_vmaf, skip=None):
        """
        Logs probes result
        :type vmaf_cq: list probe measurements (q_vmaf, q)
        :type frames: int frame count of chunk
        :type name: str chunk name
        :type skip: str None if normal results, else "high" or "low"
        :type target_q: int Calculated q to be used
        :type target_vmaf: float Calculated VMAF that would be achieved by using the q
        :return: None
        """
        if skip == "high":
            sk = " Early Skip High CQ"
        elif skip == "low":
            sk = " Early Skip Low CQ"
        else:
            sk = ""

        log(f"Chunk: {name}, Rate: {self.probing_rate}, Fr: {frames}")
        log(f"Probes: {str(sorted(vmaf_cq))[1:-1]}{sk}")
        log(f"Target Q: {target_q} VMAF: {round(target_vmaf, 2)}")

    def per_shot_target_quality(self, chunk: Chunk):
        """
        :type: Chunk chunk to probe
        :rtype: int q to use
        """
        # TODO: Refactor this mess
        vmaf_cq = []
        frames = chunk.frames

        if self.probing_rate not in (1, 2):
            self.probing_rate = self.adapt_probing_rate(self.probing_rate, frames)

        if self.probes < 3:
            return self.fast_search(chunk)

        q_list = []

        # Make middle probe
        middle_point = (self.min_q + self.max_q) // 2
        q_list.append(middle_point)
        last_q = middle_point

        score = VMAF.read_weighted_vmaf(self.vmaf_probe(chunk, last_q))
        vmaf_cq.append((score, last_q))

        # Initialize search boundary
        vmaf_lower = score
        vmaf_upper = score
        vmaf_cq_lower = last_q
        vmaf_cq_upper = last_q

        # Branch
        if score < self.target:
            next_q = self.min_q
            q_list.append(self.min_q)
        else:
            next_q = self.max_q
            q_list.append(self.max_q)

        # Edge case check
        score = VMAF.read_weighted_vmaf(self.vmaf_probe(chunk, next_q))
        vmaf_cq.append((score, next_q))

        if (next_q == self.min_q and score < self.target) or (
            next_q == self.max_q and score > self.target
        ):
            self.log_probes(
                vmaf_cq,
                frames,
                chunk.name,
                next_q,
                score,
                skip="low" if score < self.target else "high",
            )
            return next_q

        # Set boundary
        if score < self.target:
            vmaf_lower = score
            vmaf_cq_lower = next_q
        else:
            vmaf_upper = score
            vmaf_cq_upper = next_q

        # VMAF search
        for _ in range(self.probes - 2):
            new_point = self.weighted_search(
                vmaf_cq_lower, vmaf_lower, vmaf_cq_upper, vmaf_upper, self.target
            )
            if new_point in [x[1] for x in vmaf_cq]:
                break

            q_list.append(new_point)
            score = VMAF.read_weighted_vmaf(self.vmaf_probe(chunk, new_point))
            vmaf_cq.append((score, new_point))

            # Update boundary
            if score < self.target:
                vmaf_lower = score
                vmaf_cq_lower = new_point
            else:
                vmaf_upper = score
                vmaf_cq_upper = new_point

        q, q_vmaf = self.get_target_q(vmaf_cq, self.target)
        self.log_probes(vmaf_cq, frames, chunk.name, q, q_vmaf)
        # log(f'Scene_score {self.get_scene_scores(chunk, self.ffmpeg_pipe)}')
        # Plot Probes
        if self.make_plots and len(vmaf_cq) > 3:
            self.plot_probes(vmaf_cq, chunk, frames)

        return q

    def fast_search(self, chunk):
        """
        Experimental search
        Use Euler's method with known relation between cq and vmaf
        Formula -ln(1-score/100) = vmaf_cq_deriv*last_q + constant
        constant = -ln(1-score/100) - vmaf_cq_deriv*last_q
        Formula -ln(1-project.vmaf_target/100) = vmaf_cq_deriv*cq + constant
        cq = (-ln(1-project.vmaf_target/100) - constant)/vmaf_cq_deriv
        """
        vmaf_cq = []
        q_list = []

        # Make middle probe
        middle_point = (self.min_q + self.max_q) // 2
        q_list.append(middle_point)
        last_q = middle_point

        score = VMAF.read_weighted_vmaf(self.vmaf_probe(chunk, last_q))
        vmaf_cq.append((score, last_q))

        vmaf_cq_deriv = -0.18
        next_q = int(
            round(
                last_q
                + (VMAF.transform_vmaf(self.target) - VMAF.transform_vmaf(score))
                / vmaf_cq_deriv
            )
        )

        # Clamp
        if next_q < self.min_q:
            next_q = self.min_q
        if self.max_q < next_q:
            next_q = self.max_q

        # Single probe cq guess or exit to avoid divide by zero
        if self.probes == 1 or next_q == last_q:
            self.log_probes(vmaf_cq, chunk.frames, chunk.name, next_q, self.target)
            return next_q

        # Second probe at guessed value
        score_2 = VMAF.read_weighted_vmaf(self.vmaf_probe(chunk, next_q))

        # Calculate slope
        vmaf_cq_deriv = (VMAF.transform_vmaf(score_2) - VMAF.transform_vmaf(score)) / (
            next_q - last_q
        )

        # Same deal different slope
        next_q = int(
            round(
                next_q
                + (VMAF.transform_vmaf(self.target) - VMAF.transform_vmaf(score_2))
                / vmaf_cq_deriv
            )
        )

        # Clamp
        if next_q < self.min_q:
            next_q = self.min_q
        if self.max_q < next_q:
            next_q = self.max_q

        self.log_probes(vmaf_cq, chunk.frames, chunk.name, next_q, self.target)

        return next_q

    def adapt_probing_rate(self, rate, frames):
        """
        Change probing rate depending on amount of frames in scene.
        Ensure that low frame count scenes get decent amount of probes

        :param rate: given rate of probing
        :param frames: amount of frames in scene
        :return: new probing rate
        """

        # Todo: Make it depend on amount of motion in scene
        # For current moment it's 4 for everything

        if frames > 0:
            return 4

        if frames < 40:
            return 4
        elif frames < 120:
            return 8
        elif frames <= 240:
            return 10
        elif frames > 240:
            return 16

    def get_target_q(self, scores, target_quality):
        """
        Interpolating scores to get Q closest to target
        Interpolation type for 2 probes changes to linear
        """
        x = [x[1] for x in sorted(scores)]
        y = [float(x[0]) for x in sorted(scores)]

        if len(x) > 2:
            interpolation = "quadratic"
        else:
            interpolation = "linear"
        f = interpolate.interp1d(x, y, kind=interpolation)
        xnew = np.linspace(min(x), max(x), max(x) - min(x))
        tl = list(zip(xnew, f(xnew)))
        q = min(tl, key=lambda l: abs(l[1] - target_quality))

        return int(q[0]), round(q[1], 3)

    def weighted_search(self, num1, vmaf1, num2, vmaf2, target):
        """
        Returns weighted value closest to searched

        :param num1: Q of first probe
        :param vmaf1: VMAF of first probe
        :param num2: Q of second probe
        :param vmaf2: VMAF of first probe
        :param target: VMAF target
        :return: Q for new probe
        """

        dif1 = abs(VMAF.transform_vmaf(target) - VMAF.transform_vmaf(vmaf2))
        dif2 = abs(VMAF.transform_vmaf(target) - VMAF.transform_vmaf(vmaf1))

        tot = dif1 + dif2

        new_point = int(round(num1 * (dif1 / tot) + (num2 * (dif2 / tot))))
        return new_point

    def vmaf_probe(self, chunk: Chunk, q):
        """
        Calculates vmaf and returns path to json file

        :param chunk: the Chunk
        :param q: Value to make probe
        :param project: the Project
        :return : path to json file with vmaf scores
        """

        n_threads = self.n_threads if self.n_threads else self.auto_vmaf_threads()
        cmd = self.probe_cmd(
            chunk, q, self.ffmpeg_pipe, self.encoder, self.probing_rate, n_threads
        )
        pipe, utility = self.make_pipes(chunk.ffmpeg_gen_cmd, cmd)
        process_pipe(pipe, chunk, utility)
        fl = self.vmaf_runner.call_vmaf(
            chunk, self.gen_probes_names(chunk, q), vmaf_rate=self.probing_rate
        )
        return fl

    def auto_vmaf_threads(self):
        """
        Calculates number of vmaf threads based on CPU cores in system

        :return: Integer value for number of threads
        """
        cores = os.cpu_count()
        # One thread may not be enough to keep the CPU saturated, so over-provision a bit.
        over_provision_factor = 1.25
        minimum_threads = 1

        return int(max((cores / self.workers) * over_provision_factor, minimum_threads))

    def get_closest(self, q_list, q, positive=True):
        """
        Returns closest value from the list, ascending or descending

        :param q_list: list of q values that been already used
        :param q:
        :param positive: search direction, positive - only values bigger than q
        :return: q value from list
        """
        if positive:
            q_list = [x for x in q_list if x > q]
        else:
            q_list = [x for x in q_list if x < q]

        return min(q_list, key=lambda x: abs(x - q))

    def probe_cmd(
        self, chunk: Chunk, q, ffmpeg_pipe, encoder, probing_rate, n_threads
    ) -> CommandPair:
        """
        Generate and return commands for probes at set Q values
        These are specifically not the commands that are generated
        by the user or encoder defaults, since these
        should be faster than the actual encoding commands.
        These should not be moved into encoder classes at this point.
        """
        pipe = [
            "ffmpeg",
            "-y",
            "-hide_banner",
            "-loglevel",
            "error",
            "-i",
            "-",
            "-vf",
            f"select=not(mod(n\\,{probing_rate}))",
            *ffmpeg_pipe,
        ]

        probe_name = self.gen_probes_names(chunk, q).with_suffix(".ivf").as_posix()

        if encoder == "aom":
            params = [
                "aomenc",
                "--passes=1",
                f"--threads={n_threads}",
                "--tile-columns=2",
                "--tile-rows=1",
                "--end-usage=q",
                "-b",
                "8",
                "--cpu-used=6",
                f"--cq-level={q}",
                "--enable-filter-intra=0",
                "--enable-smooth-intra=0",
                "--enable-paeth-intra=0",
                "--enable-cfl-intra=0",
                "--enable-obmc=0",
                "--enable-palette=0",
                "--enable-overlay=0",
                "--enable-intrabc=0",
                "--enable-angle-delta=0",
                "--reduced-tx-type-set=1",
                "--enable-dual-filter=0",
                "--enable-intra-edge-filter=0",
                "--enable-order-hint=0",
                "--enable-flip-idtx=0",
                "--enable-dist-wtd-comp=0",
                "--enable-rect-tx=0",
                "--enable-interintra-wedge=0",
                "--enable-onesided-comp=0",
                "--enable-interintra-comp=0",
                "--enable-global-motion=0",
                "--min-partition-size=32",
                "--max-partition-size=32",
            ]
            cmd = CommandPair(pipe, [*params, "-o", probe_name, "-"])

        elif encoder == "x265":
            params = [
                "x265",
                "--log-level",
                "0",
                "--no-progress",
                "--y4m",
                "--frame-threads",
                f"{n_threads}",
                "--preset",
                "fast",
                "--crf",
                f"{q}",
            ]
            cmd = CommandPair(pipe, [*params, "-o", probe_name, "-"])

        elif encoder == "rav1e":
            params = [
                "rav1e",
                "-y",
                "-s",
                "10",
                "--threads",
                f"{n_threads}",
                "--tiles",
                "16",
                "--quantizer",
                f"{q}",
                "--low-latency",
                "--rdo-lookahead-frames",
                "5",
                "--no-scene-detection",
            ]
            cmd = CommandPair(pipe, [*params, "-o", probe_name, "-"])

        elif encoder == "vpx":
            params = [
                "vpxenc",
                "-b",
                "10",
                "--profile=2",
                "--passes=1",
                "--pass=1",
                "--codec=vp9",
                f"--threads={n_threads}",
                "--cpu-used=9",
                "--end-usage=q",
                f"--cq-level={q}",
                "--row-mt=1",
            ]
            cmd = CommandPair(pipe, [*params, "-o", probe_name, "-"])

        elif encoder == "svt_av1":
            params = [
                "SvtAv1EncApp",
                "-i",
                "stdin",
                "--lp",
                f"{n_threads}",
                "--preset",
                "8",
                "-q",
                f"{q}",
                "--tile-rows",
                "1",
                "--tile-columns",
                "2",
                "--hme",
                "0",
                "--pred-struct",
                "0",
                "--sg-filter-mode",
                "0",
                "--enable-restoration-filtering",
                "0",
                "--cdef-level",
                "0",
                "--disable-dlf",
                "0",
                "--mrp-level",
                "0",
                "--enable-tpl-la",
                "0",
                "--enable-mfmv",
                "0",
                "--enable-local-warp",
                "0",
                "--enable-global-motion",
                "0",
                "--enable-interintra-comp",
                "0",
                "--obmc-level",
                "0",
                "--rdoq-level",
                "0",
                "--filter-intra-level",
                "0",
                "--enable-intra-edge-filter",
                "0",
                "--enable-pic-based-rate-est",
                "0",
                "--pred-me",
                "0",
                "--bipred-3x3",
                "0",
                "--compound",
                "0",
                "--use-default-me-hme",
                "0",
                "--ext-block",
                "0",
                "--hbd-md",
                "0",
                "--palette-level",
                "0",
                "--umv",
                "0",
                "--tf-level",
                "3",
            ]
            cmd = CommandPair(pipe, [*params, "-b", probe_name, "-"])

        elif encoder == "svt_vp9":
            params = [
                "SvtVp9EncApp",
                "-i",
                "stdin",
                "--lp",
                f"{n_threads}",
                "-enc-mode",
                "8",
                "-q",
                f"{q}",
            ]
            # TODO: pipe needs to output rawvideo
            cmd = CommandPair(pipe, [*params, "-b", probe_name, "-"])

        elif encoder == "x264":
            params = [
                "x264",
                "--log-level",
                "error",
                "--demuxer",
                "y4m",
                "-",
                "--no-progress",
                "--threads",
                f"{n_threads}",
                "--preset",
                "medium",
                "--crf",
                f"{q}",
            ]
            cmd = CommandPair(pipe, [*params, "-o", probe_name, "-"])

        return cmd

    def search(self, q1, v1, q2, v2, target):

        if abs(target - v2) < 0.5:
            return q2

        if v1 > target and v2 > target:
            return min(q1, q2)
        if v1 < target and v2 < target:
            return max(q1, q2)

        dif1 = abs(target - v2)
        dif2 = abs(target - v1)

        tot = dif1 + dif2

        new_point = int(round(q1 * (dif1 / tot) + (q2 * (dif2 / tot))))
        return new_point

    def interpolate_data(self, vmaf_cq: list, target_quality):
        x = [x[1] for x in sorted(vmaf_cq)]
        y = [float(x[0]) for x in sorted(vmaf_cq)]

        # Interpolate data
        f = interpolate.interp1d(x, y, kind="quadratic")
        xnew = np.linspace(min(x), max(x), max(x) - min(x))

        # Getting value closest to target
        tl = list(zip(xnew, f(xnew)))
        target_quality_cq = min(tl, key=lambda l: abs(l[1] - target_quality))
        return target_quality_cq, tl, f, xnew

    def per_shot_target_quality_routine(self, chunk: Chunk):
        """
        Applies per_shot_target_quality to this chunk. Determines what the cq value should be and sets the
        per_shot_target_quality_cq for this chunk

        :param project: the Project
        :param chunk: the Chunk
        :return: None
        """
        chunk.per_shot_target_quality_cq = self.per_shot_target_quality(chunk)

    def get_scene_scores(self, chunk, ffmpeg_pipe):
        """
        Run ffmpeg scenedetection filter
        Gets average amount of motion in scene
        """

        pipecmd = [
            "ffmpeg",
            "-y",
            "-hide_banner",
            "-loglevel",
            "error",
            "-i",
            "-",
            *ffmpeg_pipe,
        ]

        params = [
            "ffmpeg",
            "-hide_banner",
            "-i",
            "-",
            "-vf",
            "select='gte(scene,0)',metadata=print",
            "-f",
            "null",
            "-",
        ]
        cmd = CommandPair(pipecmd, [*params])
        pipe, utility = self.make_pipes(chunk.ffmpeg_gen_cmd, cmd)

        history = []

        while True:
            line = pipe.stdout.readline().strip()
            if len(line) == 0 and pipe.poll() is not None:
                break
            if len(line) == 0:
                continue
            if line:
                history.append(line)

        for u_pipe in utility:
            if u_pipe.poll() is None:
                u_pipe.kill()

        if pipe.returncode != 0 and pipe.returncode != -2:
            print(f"\n:: Error in getting scene score {pipe.returncode}")
            print(f"\n:: Chunk: {chunk.index}")
            print("\n".join(history))

        scores = [x for x in history if "score" in x]

        results = []
        for x in scores:
            matches = re.findall(r"=\s*([\S\s]+)", x)
            var = float(matches[-1])
            if var < 0.3:
                results.append(var)

        result = round(np.average(results), 4) * 1000
        return result

    def gen_probes_names(self, chunk: Chunk, q):
        """
        Make name of vmaf probe
        """
        return chunk.fake_input_path.with_name(f"v_{q}{chunk.name}").with_suffix(".ivf")

    def make_pipes(self, ffmpeg_gen_cmd: Command, command: CommandPair):

        ffmpeg_gen_pipe = subprocess.Popen(
            ffmpeg_gen_cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )

        ffmpeg_pipe = subprocess.Popen(
            command[0],
            stdin=ffmpeg_gen_pipe.stdout,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

        pipe = subprocess.Popen(
            command[1],
            stdin=ffmpeg_pipe.stdout,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )

        utility = (ffmpeg_gen_pipe, ffmpeg_pipe)
        return pipe, utility

    def plot_probes(self, vmaf_cq, chunk: Chunk, frames):
        """
        Makes graph with probe decisions
        """
        if plt is None:
            log(
                "Matplotlib is not installed or could not be loaded\
            . Unable to plot probes."
            )
            return
        # Saving plot of vmaf calculation

        x = [x[1] for x in sorted(vmaf_cq)]
        y = [float(x[0]) for x in sorted(vmaf_cq)]

        cq, tl, f, xnew = self.interpolate_data(vmaf_cq, self.target)
        matplotlib.use("agg")
        plt.ioff()
        plt.plot(xnew, f(xnew), color="tab:blue", alpha=1)
        plt.plot(x, y, "p", color="tab:green", alpha=1)
        plt.plot(cq[0], cq[1], "o", color="red", alpha=1)
        plt.grid(True)
        plt.xlim(self.min_q, self.max_q)
        vmafs = [int(x[1]) for x in tl if isinstance(x[1], float) and not isnan(x[1])]
        plt.ylim(min(vmafs), max(vmafs) + 1)
        plt.ylabel("VMAF")
        plt.title(f"Chunk: {chunk.name}, Frames: {frames}")
        plt.xticks(np.arange(self.min_q, self.max_q + 1, 1.0))
        temp = self.temp / chunk.name
        plt.savefig(f"{temp}.png", dpi=200, format="png")
        plt.close()

    def make_q_file(self, q_list, chunk):
        qfile = chunk.fake_input_path.with_name(f"probe_{chunk.name}").with_suffix(
            ".txt"
        )
        with open(qfile, "w") as fl:
            text = ""

            for x in q_list:
                text += str(x) + "\n"
            fl.write(text)
        return qfile

    def per_frame_probe_cmd(
        self, chunk: Chunk, q, encoder, probing_rate, qp_file
    ) -> CommandPair:
        """
        Generate and return commands for probes at set Q values
        These are specifically not the commands that are generated
        by the user or encoder defaults, since these
        should be faster than the actual encoding commands.
        These should not be moved into encoder classes at this point.
        """
        pipe = [
            "ffmpeg",
            "-y",
            "-hide_banner",
            "-loglevel",
            "error",
            "-i",
            "-",
            "-vf",
            f"select=not(mod(n\\,{probing_rate}))",
            *self.ffmpeg_pipe,
        ]

        probe_name = self.gen_probes_names(chunk, q).with_suffix(".ivf").as_posix()
        if encoder == "svt_av1":
            params = [
                "SvtAv1EncApp",
                "-i",
                "stdin",
                "--preset",
                "8",
                "--rc",
                "0",
                "--passes",
                "1",
                "--use-q-file",
                "1",
                "--qpfile",
                f"{qp_file.as_posix()}",
            ]

            cmd = CommandPair(pipe, [*params, "-b", probe_name, "-"])
        else:
            print("supported only by SVT-AV1")
            exit()
        """
        elif encoder == 'x265':
            params = [
                'x265', '--log-level', '0', '--no-progress', '--y4m', '--preset',
                'fast', '--crf', f'{q}'
            ]
            cmd = CommandPair(pipe, [*params, '-o', probe_name, '-'])
        """

        return cmd

    def per_frame_probe(self, q_list, q, chunk):
        qfile = chunk.make_q_file(q_list)
        cmd = self.per_frame_probe_cmd(chunk, q, self.encoder, 1, qfile)
        pipe, utility = self.make_pipes(chunk.ffmpeg_gen_cmd, cmd)
        process_pipe(pipe, chunk, utility)
        fl = self.vmaf_runner.call_vmaf(chunk, self.gen_probes_names(chunk, q))
        jsn = VMAF.read_json(fl)
        vmafs = [x["metrics"]["vmaf"] for x in jsn["frames"]]
        return vmafs

    def add_probes_to_frame_list(self, frame_list, q_list, vmafs):
        frame_list = list(frame_list)
        for index, q_vmaf in enumerate(zip(q_list, vmafs)):
            frame_list[index]["probes"].append((q_vmaf[0], q_vmaf[1]))

        return frame_list

    def per_frame_target_quality(self, chunk):
        frames = chunk.frames
        frame_list = [{"frame_number": x, "probes": []} for x in range(frames)]

        for _ in range(self.probes):
            q_list = self.gen_next_q(frame_list, chunk)
            vmafs = self.per_frame_probe(q_list, 1, chunk)
            frame_list = self.add_probes_to_frame_list(frame_list, q_list, vmafs)
            mse = round(
                self.get_square_error(
                    [x["probes"][-1][1] for x in frame_list], self.target
                ),
                2,
            )
            # print(':: MSE:', mse)

            if mse < 1.0:
                return q_list

        return q_list

    def get_square_error(self, ls, target):
        total = 0
        for i in ls:
            dif = i - target
            total += dif ** 2
        mse = total / len(ls)
        return mse

    def gen_next_q(self, frame_list, chunk):
        q_list = []

        probes = len(frame_list[0]["probes"])

        if probes == 0:
            return [self.min_q] * len(frame_list)
        elif probes == 1:
            return [self.max_q] * len(frame_list)
        else:
            for probe in frame_list:

                x = [x[0] for x in probe["probes"]]
                y = [x[1] for x in probe["probes"]]

                if probes > 2:
                    if len(x) != len(set(x)):
                        q_list.append(probe["probes"][-1][0])
                        continue

                interpolation = "quadratic" if probes > 2 else "linear"

                f = interpolate.interp1d(x, y, kind=interpolation)
                xnew = np.linspace(min(x), max(x), max(x) - min(x))
                tl = list(zip(xnew, f(xnew)))
                q = min(tl, key=lambda l: abs(l[1] - self.target))

                q_list.append(int(round(q[0])))

            return q_list


"""
def frame_types_probe(chunk: Chunk, q, ffmpeg_pipe, encoder, probing_rate, qp_file) -> CommandPair:

    probe_name = gen_probes_names(chunk, q).with_suffix('.ivf').as_posix()

    pipe = ['ffmpeg', '-y', '-hide_banner', '-loglevel', 'error', '-i', '-', '-vf', f'select=not(mod(n\\,{probing_rate}))',
            *ffmpeg_pipe]

    params = ['x265', '--log-level', '0', '--no-progress',
              '--y4m', '--preset', 'fast', '--crf', f'{q}']

    cmd = CommandPair(pipe, [*params, '-o', probe_name, '-'])

    return cmd
"""



import json
import shutil
import sys
import time
from multiprocessing.managers import BaseManager
from pathlib import Path
from typing import List

from av1an.chunk import Chunk
from av1an.chunk.chunk_queue import load_or_gen_chunk_queue
from av1an.ffmpeg import extract_audio
from av1an.fp_reuse import segment_first_pass
from av1an.logger import log, set_log
from av1an.project.Project import Project
from av1an.split import split_routine
from av1an.startup.file_validation import process_inputs

from av1an.utils import frame_probe, terminate
from av1an.vmaf import VMAF

from .Counter import BaseManager, Counter, Manager
from .Queue import Queue


class Main:
    def __init__(self, args):
        self.file_queue: list[Path] = []
        self.args = args
        self.file_queue = process_inputs(args.input)
        self.projects = self.create_project_list()

    def create_project_list(self):
        """
        Returns list of initialized Project objects with single input
        """
        queue = []
        for file in self.file_queue:
            project = Project(vars(self.args))
            project.input = file
            project.outputs_filenames()
            queue.append(project)
        return queue

    def run(self):
        """
        Run encoding in queue or single file
        """
        for i, proj in enumerate(self.projects):
            if proj.output_file.exists() and len(self.projects) > 1:
                print(
                    f":: Skipping file {proj.input.name}\n:: Outputfile {proj.output_file.name} exists"
                )

                # Don't print new line on last project to console
                if i + 1 < len(self.projects):
                    print()

                continue
            try:
                tm = time.time()

                if len(self.projects) > 1:
                    print(f":: Encoding file {proj.input.name}")
                EncodingManager().encode_file(proj)

                print(f"Finished: {round(time.time() - tm, 1)}s\n")
            except KeyboardInterrupt:
                print("Encoding stopped")
                sys.exit()


class EncodingManager:
    def __init__(self):
        self.workers = None
        self.vmaf = None
        self.initial_frames = 0

    def encode_file(self, project: Project):
        """
        Encodes a single video file on the local machine.

        :param project: The project for this encode
        :return: None
        """

        project.setup()
        set_log(project.logging, project.temp)

        # find split locations
        split_locations = split_routine(project, project.resume)

        # create a chunk queue
        chunk_queue = load_or_gen_chunk_queue(project, project.resume, split_locations)

        self.done_file(project, chunk_queue)
        if not project.resume:
            extract_audio(project.input, project.temp, project.audio_params)

            if project.reuse_first_pass:
                segment_first_pass(project.temp, split_locations)

        # do encoding loop
        project.determine_workers()
        self.startup(project, chunk_queue)
        queue = Queue(project, chunk_queue)
        queue.encoding_loop()

        if queue.status.lower() == "fatal":
            msg = "FATAL Encoding process encountered fatal error, shutting down"
            print("\n::", msg)
            log(msg)
            sys.exit(1)

        # concat
        project.concat_routine()

        if project.vmaf or project.vmaf_plots:
            self.vmaf = VMAF(
                n_threads=project.n_threads,
                model=project.vmaf_path,
                res=project.vmaf_res,
                vmaf_filter=project.vmaf_filter,
            )
            self.vmaf.plot_vmaf(project.input, project.output_file, project)

        # Delete temp folders
        if not project.keep:
            shutil.rmtree(project.temp)

    def done_file(self, project: Project, chunk_queue: List[Chunk]):
        done_path = project.temp / "done.json"
        if project.resume and done_path.exists():
            log("Resuming...")
            with open(done_path) as done_file:
                data = json.load(done_file)

            project.set_frames(data["frames"])
            done = len(data["done"])
            self.initial_frames = sum(data["done"].values())
            log(f"Resumed with {done} encoded clips done")
        else:
            self.initial_frames = 0
            total = project.get_frames()
            d = {"frames": total, "done": {}}
            with open(done_path, "w") as done_file:
                json.dump(d, done_file)

    def startup(self, project: Project, chunk_queue: List[Chunk]):
        clips = len(chunk_queue)
        project.workers = min(project.workers, clips)
        print(
            f"\rQueue: {clips} Workers: {project.workers} Passes: {project.passes}\n"
            f'Params: {" ".join(project.video_params)}'
        )
        BaseManager.register("Counter", Counter)
        counter = Manager().Counter(
            project.get_frames(), self.initial_frames, not project.quiet
        )
        project.counter = counter



import os
import re

from av1an.project import Project
from av1an.chunk import Chunk
from av1an.commandtypes import MPCommands, CommandPair, Command
from .encoder import Encoder
from av1an.utils import list_index_of_regex, terminate


class Vpx(Encoder):
    def __init__(self):
        super().__init__(
            encoder_bin="vpxenc",
            encoder_help="vpxenc --help",
            default_args=[
                "--codec=vp9",
                "-b",
                "10",
                "--profile=2",
                "--threads=4",
                "--cpu-used=0",
                "--end-usage=q",
                "--cq-level=30",
                "--row-mt=1",
            ],
            default_passes=2,
            default_q_range=(15, 55),
            output_extension="ivf",
        )

    def compose_1_pass(self, a: Project, c: Chunk, output: str) -> MPCommands:
        return [
            CommandPair(
                Encoder.compose_ffmpeg_pipe(a),
                ["vpxenc", "--passes=1", *a.video_params, "-o", output, "-"],
            )
        ]

    def compose_2_pass(self, a: Project, c: Chunk, output: str) -> MPCommands:
        return [
            CommandPair(
                Encoder.compose_ffmpeg_pipe(a),
                [
                    "vpxenc",
                    "--passes=2",
                    "--pass=1",
                    *a.video_params,
                    f"--fpf={c.fpf}",
                    "-o",
                    os.devnull,
                    "-",
                ],
            ),
            CommandPair(
                Encoder.compose_ffmpeg_pipe(a),
                [
                    "vpxenc",
                    "--passes=2",
                    "--pass=2",
                    *a.video_params,
                    f"--fpf={c.fpf}",
                    "-o",
                    output,
                    "-",
                ],
            ),
        ]

    def man_q(self, command: Command, q: int) -> Command:
        """Return command with new cq value

        :param command: old command
        :param q: new cq value
        :return: command with new cq value"""

        adjusted_command = command.copy()

        i = list_index_of_regex(adjusted_command, r"--cq-level=.+")
        adjusted_command[i] = f"--cq-level={q}"

        return adjusted_command

    def match_line(self, line: str):
        """Extract number of encoded frames from line.

        :param line: one line of text output from the encoder
        :return: match object from re.search matching the number of encoded frames"""

        if "fatal" in line.lower():
            print("\n\nERROR IN ENCODING PROCESS\n\n", line)
            terminate()
        if "Pass 2/2" in line or "Pass 1/1" in line:
            return re.search(r"frame.*?/([^ ]+?) ", line)



#!/bin/env python
import argparse
import sys
from pathlib import Path
from .project import Project


class Args:
    """
    Class responsible for arg parsing
    Creation of original project file
    """

    def __init__(self):
        self.parser = self.arg_parsing()
        self.defaults = self.get_defaults()
        self.parsed = None
        self.project = None

    def get_project(self):
        """
        Create and return project object with all parameters
        """
        if not self.parsed:
            self.parse()

        self.project = Project(self.parsed)

        if self.project.config:
            self.save_load_project_file()

        return self.project

    def get_defaults(self) -> dict:
        """
        Get dictionary of default values specified in arg_parsing()
        """
        return vars(self.parser.parse_args([]))

    def get_difference(self) -> dict:
        """
        Return difference of defaults and new
        """
        return dict([x for x in self.parsed.items() if x not in self.defaults.items()])

    def parse(self):
        """
        Parse command line parameters provided by user
        """
        self.parsed = vars(self.parser.parse_args())
        if not self.parsed["input"]:
            self.parser.print_help()
            sys.exit()

    def save_load_project_file(self):
        """
        Saves current/Loads given project file, loads saved project first and when overwrites only unique values from current parse
        """
        cfg_path = Path(self.project.config)

        if cfg_path.exists():

            new = self.get_difference()
            self.project.load_project_from_file(self.project.config)
            self.project.load_project(new)

        else:
            self.project.save_project_to_file(self.project.config)

    def arg_parsing(self):
        """Command line parsing and setting default variables"""
        parser = argparse.ArgumentParser()

        # Input/Output/Temp
        io_group = parser.add_argument_group("Input and Output")
        io_group.add_argument(
            "--input",
            "-i",
            nargs="+",
            type=Path,
            help="Input File or Vapoursynth (.py,.vpy) script",
        )
        io_group.add_argument(
            "--temp", type=str, default=None, help="Set temp folder path"
        )
        io_group.add_argument(
            "--output_file", "-o", type=str, default=None, help="Specify output file"
        )
        io_group.add_argument(
            "--mkvmerge",
            help="Use mkvmerge instead of ffmpeg to concatenate",
            action="store_true",
        )

        io_group.add_argument(
            "--quiet",
            "-q",
            help="Disable printing tqdm and scenedetect data to terminal",
            action="store_true",
        )
        io_group.add_argument(
            "--logging", "-log", type=str, default=None, help="Enable logging"
        )
        io_group.add_argument(
            "--resume", "-r", help="Resuming previous session", action="store_true"
        )
        io_group.add_argument(
            "--keep", help="Keep temporally folder after encode", action="store_true"
        )
        io_group.add_argument(
            "--config",
            "-c",
            type=str,
            default=None,
            help="Path to config file, create if doesn't exists",
        )
        io_group.add_argument(
            "--webm", default=False, action="store_true", help="Output to webm"
        )

        # Splitting
        split_group = parser.add_argument_group("Splitting")
        split_group.add_argument(
            "--chunk_method",
            "-cm",
            type=str,
            default=None,
            help="Method for creating chunks",
            choices=["select", "vs_ffms2", "vs_lsmash", "hybrid"],
        )
        split_group.add_argument(
            "--scenes", "-s", type=str, default=None, help="File location for scenes"
        )
        split_group.add_argument(
            "--split_method",
            type=str,
            default="pyscene",
            help="Specify splitting method",
            choices=["none", "pyscene", "aom_keyframes", "ffmpeg"],
        )
        split_group.add_argument(
            "--extra_split",
            "-xs",
            type=int,
            default=240,
            help="Number of frames after which make split",
        )

        # PySceneDetect split
        split_group.add_argument(
            "--threshold", "-tr", type=float, default=35, help="PySceneDetect Threshold"
        )
        split_group.add_argument(
            "--min_scene_len",
            type=int,
            default=60,
            help="Minimum number of frames in a split",
        )

        # AOM Keyframe split
        split_group.add_argument(
            "--reuse_first_pass",
            help="Reuse the first pass from aom_keyframes split on the chunks",
            action="store_true",
        )

        # Encoding
        encode_group = parser.add_argument_group("Encoding")
        encode_group.add_argument(
            "--passes",
            "-p",
            type=int,
            default=None,
            help="Specify encoding passes",
            choices=[1, 2],
        )
        encode_group.add_argument(
            "--video_params", "-v", type=str, default=None, help="encoding settings"
        )
        encode_group.add_argument(
            "--encoder",
            "-enc",
            type=str,
            default="aom",
            help="Choosing encoder",
            choices=[
                "aom",
                "svt_av1",
                "svt_vp9",
                "rav1e",
                "vpx",
                "x265",
                "x264",
                "vvc",
            ],
        )
        encode_group.add_argument(
            "--workers", "-w", type=int, default=0, help="Number of workers"
        )
        encode_group.add_argument(
            "--no_check", "-n", help="Do not check encodings", action="store_true"
        )
        encode_group.add_argument(
            "--force",
            help="Force encoding if input args seen as invalid",
            action="store_true",
        )

        # VVC
        encode_group.add_argument(
            "--vvc_conf", type=Path, default=None, help="Path to VVC confing file"
        )

        # FFmpeg params
        ffmpeg_group = parser.add_argument_group("FFmpeg")
        ffmpeg_group.add_argument(
            "--ffmpeg", "-ff", type=str, default="", help="FFmpeg commands"
        )
        ffmpeg_group.add_argument(
            "--audio_params",
            "-a",
            type=str,
            default="-c:a copy",
            help="FFmpeg audio settings",
        )
        ffmpeg_group.add_argument(
            "--pix_format",
            "-fmt",
            type=str,
            default="yuv420p10le",
            help="FFmpeg pixel format",
        )

        # Vmaf
        vmaf_group = parser.add_argument_group("VMAF")
        vmaf_group.add_argument(
            "--vmaf", help="Calculating vmaf after encode", action="store_true"
        )
        vmaf_group.add_argument(
            "--vmaf_path", type=Path, default=None, help="Path to vmaf models"
        )
        vmaf_group.add_argument(
            "--vmaf_res",
            type=str,
            default="1920x1080",
            help="Resolution used in vmaf calculation",
        )
        vmaf_group.add_argument(
            "--n_threads", type=int, default=None, help="Threads for vmaf calculation"
        )

        # Target Quality
        tq_group = parser.add_argument_group("Target Quality")
        tq_group.add_argument("--target_quality", type=float, help="Value to target")
        tq_group.add_argument(
            "--target_quality_method",
            type=str,
            default="per_shot",
            help="Method selection for target quality",
            choices=["per_frame", "per_shot"],
        )
        tq_group.add_argument(
            "--probes",
            type=int,
            default=4,
            help="Number of probes to make for target_quality",
        )
        tq_group.add_argument(
            "--min_q", type=int, default=None, help="Min q for target_quality"
        )
        tq_group.add_argument(
            "--max_q", type=int, default=None, help="Max q for target_quality"
        )
        tq_group.add_argument(
            "--vmaf_plots",
            help="Make plots of probes in temp folder",
            action="store_true",
        )
        tq_group.add_argument(
            "--probing_rate",
            type=int,
            default=4,
            help="Framerate for probes, 0 - original",
        )
        tq_group.add_argument(
            "--vmaf_filter",
            type=str,
            default=None,
            help="Filter applied to source at vmaf calcualation, use if you crop source",
        )

        # Misc
        misc_group = parser.add_argument_group("Misc")
        misc_group.add_argument(
            "--version", action="version", version=f"Av1an version: 6.1-2"
        )
        # Initialize project with initial values
        return parser



import time
import sys
import concurrent
import concurrent.futures
from typing import List
from av1an.target_quality import TargetQuality
from av1an.encoder import ENCODERS
from av1an.utils import frame_probe, terminate
from av1an.resume import write_progress_file
from av1an.chunk import Chunk
from av1an.logger import log
from pathlib import Path
from .Pipes import tqdm_bar


class Queue:
    """
    Queue manager with ability to add/remove/restart jobs
    """

    def __init__(self, project, chunk_queue):
        self.chunk_queue = chunk_queue
        self.queue = []
        self.project = project
        self.thread_executor = concurrent.futures.ThreadPoolExecutor()
        self.status = "Ok"
        self.tq = TargetQuality(project) if project.target_quality else None

    def encoding_loop(self):
        if len(self.chunk_queue) != 0:
            with concurrent.futures.ThreadPoolExecutor(
                max_workers=self.project.workers
            ) as executor:
                future_cmd = {
                    executor.submit(self.encode_chunk, cmd): cmd
                    for cmd in self.chunk_queue
                }
                for future in concurrent.futures.as_completed(future_cmd):
                    try:
                        future.result()
                    except Exception as exc:
                        _, _, exc_tb = sys.exc_info()
                        print(f"Encoding error {exc}\nAt line {exc_tb.tb_lineno}")
                        terminate()
        self.project.counter.close()

    def encode_chunk(self, chunk: Chunk):
        """
        Encodes a chunk. If chunk fails, restarts it limited amount of times.
        Return if executed just fine, sets status fatal for queue if failed

        :param chunk: The chunk to encode
        :return: None
        """
        restart_count = 0

        while restart_count < 3:
            try:
                st_time = time.time()

                chunk_frames = chunk.frames

                log(f"Enc: {chunk.index}, {chunk_frames} fr")

                # Target Quality Mode
                if self.project.target_quality:
                    if self.project.target_quality_method == "per_shot":
                        self.tq.per_shot_target_quality_routine(chunk)
                    if self.project.target_quality_method == "per_frame":
                        self.tq.per_frame_target_quality_routine(chunk)

                # skip first pass if reusing
                start = (
                    2
                    if self.project.reuse_first_pass and self.project.passes >= 2
                    else 1
                )

                # Run all passes for this chunk
                for current_pass in range(start, self.project.passes + 1):
                    tqdm_bar(
                        self.project,
                        chunk,
                        self.project.encoder,
                        self.project.counter,
                        chunk_frames,
                        self.project.passes,
                        current_pass,
                    )

                # get the number of encoded frames, if no check assume it worked and encoded same number of frames
                encoded_frames = (
                    chunk_frames
                    if self.project.no_check
                    else self.frame_check_output(chunk, chunk_frames)
                )

                # write this chunk as done if it encoded correctly
                if encoded_frames == chunk_frames:
                    write_progress_file(
                        Path(self.project.temp / "done.json"), chunk, encoded_frames
                    )

                enc_time = round(time.time() - st_time, 2)
                log(f"Done: {chunk.index} Fr: {encoded_frames}/{chunk_frames}")
                log(f"Fps: {round(encoded_frames / enc_time, 4)} Time: {enc_time} sec.")
                return

            except Exception as e:
                msg1, msg2, msg3 = (
                    f"Chunk #{chunk.index} crashed",
                    f"Exception: {type(e)} {e}",
                    "Restarting chunk",
                )
                log(msg1, msg2, msg3)
                print(f"{msg1}\n::{msg2}\n::{msg3}")
                restart_count += 1

        msg1, msg2 = (
            "FATAL",
            f"Chunk #{chunk.index} failed more than 3 times, shutting down thread",
        )
        log(msg1, msg2)
        print(f"::{msg1}\n::{msg2}")
        self.status = "FATAL"

    def frame_check_output(
        self, chunk: Chunk, expected_frames: int, last_chunk=False
    ) -> int:
        actual_frames = frame_probe(chunk.output_path)
        if actual_frames != expected_frames:
            msg = f"Chunk #{chunk.index}: {actual_frames}/{expected_frames} fr"
            log(msg)
            print("::" + msg)
        return actual_frames



import re
import subprocess
import sys
from subprocess import Popen

from av1an.logger import log
from av1an.vapoursynth import compose_vapoursynth_pipe

if sys.platform == "linux":
    from os import mkfifo


def ffmpeg(video, threshold, min_scene_len, total_frames, is_vs, temp):
    """
    Running FFMPEG detection on source video for segmenting.
    Usually the optimal threshold is 0.1 - 0.3 but it can vary a lot
    based on your source content.

    Threshold value increased by x100 for matching with pyscene range
    """

    log(f"Starting FFMPEG detection:")
    log(f"Threshold: {threshold}")
    log(f"Is Vapoursynth input: {is_vs}")
    scenes = []
    frame: int = 0

    if is_vs:
        if sys.platform == "linux":
            vspipe_fifo = temp / "vspipe.y4m"
            mkfifo(vspipe_fifo)
        else:
            vspipe_fifo = None

        vspipe_cmd = compose_vapoursynth_pipe(video, vspipe_fifo)
        vspipe_process = Popen(vspipe_cmd)

    cmd = [
        "ffmpeg",
        "-hwaccel",
        "auto",
        "-hide_banner",
        "-i",
        str(vspipe_fifo if is_vs else video.as_posix()),
        "-an",
        "-sn",
        "-vf",
        "scale='min(960,iw):-1:flags=neighbor',select='gte(scene,0)',metadata=print",
        "-f",
        "null",
        "-",
    ]
    pipe = Popen(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True
    )

    while True:
        line = pipe.stdout.readline().strip()
        if len(line) == 0 and pipe.poll() is not None:
            break
        if len(line) == 0:
            continue

        if "frame" in line:
            match = re.findall(r":(\d+)", line)
            if match:
                frame = int(match[0])
                continue

        if "score" in line:
            matches = re.findall(r"=\s*([\S\s]+)", line)
            if matches:
                score = float(matches[-1]) * 100
                if score > threshold and frame - max(scenes, default=0) > min_scene_len:
                    scenes.append(frame)

    if pipe.returncode != 0 and pipe.returncode != -2:
        print(f"\n:: Error in ffmpeg scenedetection {pipe.returncode}")
        print("\n".join(scenes))

    if is_vs:
        vspipe_process.wait()

    log(f"Found split points: {len(scenes)}\n")
    log(f"Splits: {scenes}\n")

    return scenes


from multiprocessing.managers import BaseManager

try:
    from tqdm import tqdm
except ImportError:
    tqdm = None


def Manager():
    """
    Thread save manager for frame counter
    """
    m = BaseManager()
    m.start()
    return m


class Counter:
    """
    Frame Counter
    """

    def __init__(self, total, initial, use_tqdm=True):
        self.first_update = True
        self.initial = initial
        self.left = total - initial
        self.current = 0
        self.use_tqdm = use_tqdm and (tqdm is not None)
        if use_tqdm:
            self.tqdm_bar = tqdm(
                total=self.left,
                initial=0,
                dynamic_ncols=True,
                unit="fr",
                leave=True,
                smoothing=0.01,
            )

    def update(self, value):
        if self.use_tqdm:
            if self.first_update:
                self.tqdm_bar.reset(self.left)
                self.first_update = False
            self.tqdm_bar.update(value)
        else:
            self.current += value

    def close(self):
        if self.use_tqdm:
            self.tqdm_bar.close()

    def get_frames(self):
        return self.current



#!/bin/env python

import re
import subprocess
from pathlib import Path
from subprocess import PIPE, STDOUT
from typing import List
from av1an.logger import log

# TODO: redo to module, add ffmpeg scenedetection for fallback


def frame_probe_ffmpeg(source: Path):
    """
    Get frame count.
    Direct counting of frame count using ffmpeg. Slow, Precise.
    :param: source: Path to input file
    """
    cmd = [
        "ffmpeg",
        "-hide_banner",
        "-i",
        source.as_posix(),
        "-map",
        "0:v:0",
        "-f",
        "null",
        "-",
    ]
    r = subprocess.run(cmd, stdout=PIPE, stderr=PIPE)
    matches = re.findall(
        r"frame=\s*([0-9]+)\s", r.stderr.decode("utf-8") + r.stdout.decode("utf-8")
    )
    return int(matches[-1])


def get_frametypes(file: Path) -> List:
    """
    Read file and return list with all frame types
    :param file: Path for file
    :return: list with sequence of frame types
    """

    frames = []

    ff = [
        "ffmpeg",
        "-hide_banner",
        "-i",
        file.as_posix(),
        "-vf",
        "showinfo",
        "-f",
        "null",
        "-loglevel",
        "debug",
        "-",
    ]

    pipe = subprocess.Popen(ff, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

    while True:
        line = pipe.stdout.readline().strip().decode("utf-8")

        if len(line) == 0 and pipe.poll() is not None:
            break

        frames.append(line)

    return frames


def get_keyframes(file: Path) -> List[int]:
    """
    Read file info and return list of all keyframes

    :param file: Path for input file
    :return: list with frame numbers of keyframes
    """

    keyframes = []

    ff = [
        "ffmpeg",
        "-hide_banner",
        "-i",
        file.as_posix(),
        "-vf",
        r"select=eq(pict_type\,PICT_TYPE_I)",
        "-f",
        "null",
        "-loglevel",
        "debug",
        "-",
    ]

    pipe = subprocess.Popen(ff, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

    while True:
        line = pipe.stdout.readline().strip().decode("utf-8")

        if len(line) == 0 and pipe.poll() is not None:
            break

        match = re.search(r"n:([0-9]+)\.[0-9]+ pts:.+key:1", line)
        if match:
            keyframe = int(match.group(1))
            keyframes.append(keyframe)

    return keyframes


def extract_audio(input_vid: Path, temp, audio_params):
    """Extracting audio from source, transcoding if needed."""
    log(f"Audio processing")
    log(f'Params: {" ".join(audio_params)}')
    audio_file = temp / "audio.mkv"

    # Checking is source have audio track
    check = [
        "ffmpeg",
        "-y",
        "-hide_banner",
        "-loglevel",
        "error",
        "-ss",
        "0",
        "-i",
        input_vid.as_posix(),
        "-t",
        "0",
        "-vn",
        "-c:a",
        "copy",
        "-f",
        "null",
        "-",
    ]
    is_audio_here = len(subprocess.run(check, stdout=PIPE, stderr=STDOUT).stdout) == 0

    # If source have audio track - process it
    if is_audio_here:
        cmd = (
            "ffmpeg",
            "-y",
            "-hide_banner",
            "-loglevel",
            "error",
            "-i",
            input_vid.as_posix(),
            "-map_metadata",
            "-1",
            "-dn",
            "-vn",
            *audio_params,
            audio_file.as_posix(),
        )
        subprocess.run(cmd)



from .arg_parse import Args
from .manager import Manager
from .startup.setup import startup_check


def main():
    """
    Running Av1an CLI
    """
    parser = Args()
    project = parser.get_project()
    startup_check(project)
    manager = Manager.Main(project)
    manager.run()


if __name__ == "__main__":
    main()


import os
import re

from av1an.project import Project
from av1an.chunk import Chunk
from av1an.commandtypes import MPCommands, CommandPair, Command
from av1an.encoder.encoder import Encoder
from av1an.utils import list_index_of_regex, terminate


class SvtAv1(Encoder):
    def __init__(self):
        super(SvtAv1, self).__init__(
            encoder_bin="SvtAv1EncApp",
            encoder_help="SvtAv1EncApp --help",
            default_args=["--preset", "4", "--rc", "0", "-q", "25"],
            default_passes=1,
            default_q_range=(15, 50),
            output_extension="ivf",
        )

    def compose_1_pass(self, a: Project, c: Chunk, output: str) -> MPCommands:
        return [
            CommandPair(
                Encoder.compose_ffmpeg_pipe(a),
                [
                    "SvtAv1EncApp",
                    "-i",
                    "stdin",
                    "--progress",
                    "2",
                    *a.video_params,
                    "-b",
                    output,
                    "-",
                ],
            )
        ]

    def compose_2_pass(self, a: Project, c: Chunk, output: str) -> MPCommands:
        return [
            CommandPair(
                Encoder.compose_ffmpeg_pipe(a),
                [
                    "SvtAv1EncApp",
                    "-i",
                    "stdin",
                    "--progress",
                    "2",
                    "--irefresh-type",
                    "2",
                    *a.video_params,
                    "--pass",
                    "1",
                    "--stats",
                    f"{c.fpf}.stat",
                    "-b",
                    os.devnull,
                    "-",
                ],
            ),
            CommandPair(
                Encoder.compose_ffmpeg_pipe(a),
                [
                    "SvtAv1EncApp",
                    "-i",
                    "stdin",
                    "--progress",
                    "2",
                    "--irefresh-type",
                    "2",
                    *a.video_params,
                    "--pass",
                    "2",
                    "--stats",
                    f"{c.fpf}.stat",
                    "-b",
                    output,
                    "-",
                ],
            ),
        ]

    def mod_command(self, command, chunk) -> Command:
        """Return new command with q_file

        :param command: old command
        :param q: q list
        :return: command with new cq value"""
        adjusted_command = command.copy()

        i = list_index_of_regex(adjusted_command, r"(--qp|-q)")

        qp_file = chunk.make_q_file(chunk.per_frame_target_quality_q_list)

        new = ["--use-q-file", "1", "--qpfile", f"{qp_file.as_posix()}"]
        new_cmd = adjusted_command[:i] + new + adjusted_command[i + 2 :]

        return new_cmd

    def man_q(self, command: Command, q: int) -> Command:
        """Return command with new cq value

        :param command: old command
        :param q: new cq value
        :return: command with new cq value"""

        adjusted_command = command.copy()

        i = list_index_of_regex(adjusted_command, r"(--qp|-q)")
        adjusted_command[i + 1] = f"{q}"

        return adjusted_command

    def match_line(self, line: str):
        """Extract number of encoded frames from line.

        :param line: one line of text output from the encoder
        :return: match object from re.search matching the number of encoded frames"""
        if "error" in line.lower():
            print("\n\nERROR IN ENCODING PROCESS\n\n", line)
            terminate()
        return re.search(r"Encoding frame\s+(\d+)", line)



import setuptools

# TODO: rewrite it in rust

REQUIRES = [
    "numpy",
    "scenedetect[opencv]",
    "opencv-python",
    "tqdm",
    "psutil",
    "scipy",
    "matplotlib",
]

with open("README.md", "r") as f:
    long_description = f.read()

version = "6.1-3"

setuptools.setup(
    name="Av1an",
    version=version,
    author="Master_Of_Zen",
    author_email="master_of_zen@protonmail.com",
    description="Cross-platform command-line AV1 / VP9 / HEVC / H264 / VVC encoding framework with per scene quality encoding",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/master-of-zen/Av1an",
    packages=setuptools.find_packages(".", exclude="tests"),
    install_requires=REQUIRES,
    py_modules=["av1an"],
    entry_points={"console_scripts": ["av1an=av1an:main"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



import json
import sys
import os
import shutil
from psutil import virtual_memory
from distutils.spawn import find_executable
from pathlib import Path
from av1an.commandtypes import Command
from av1an.utils import frame_probe_fast, hash_path, terminate
from av1an.concat import vvc_concat, concatenate_ffmpeg, concatenate_mkvmerge
from av1an.logger import log
from av1an.vapoursynth import create_vs_file, frame_probe_vspipe


class Project(object):
    def __init__(self, initial_data):

        # Project info
        self.frames: int = 0
        self.counter = None
        self.is_vs: bool = None

        # Input/Output/Temp
        self.input: Path = None
        self.temp: Path = None
        self.output_file: Path = None
        self.mkvmerge: bool = None
        self.config = None
        self.webm = None

        # Splitting
        self.chunk_method: str = None
        self.scenes: Path = None
        self.split_method: str = None
        self.extra_split: int = None
        self.min_scene_len: int = None

        # PySceneDetect split
        self.threshold: float = None

        # AOM Keyframe split
        self.reuse_first_pass: bool = None

        # Encoding
        self.passes = None
        self.video_params: Command = None
        self.encoder: str = None
        self.workers: int = None

        # FFmpeg params
        self.ffmpeg_pipe: Command = None
        self.ffmpeg: str = None
        self.audio_params = None
        self.pix_format: Command = None

        # Misc
        self.quiet = False
        self.logging = None
        self.resume: bool = None
        self.no_check: bool = None
        self.keep: bool = None
        self.force: bool = None

        # Vmaf
        self.vmaf: bool = None
        self.vmaf_path: str = None
        self.vmaf_res: str = None

        # Target Quality
        self.target_quality: int = None
        self.probes: int = None
        self.min_q: int = None
        self.max_q: int = None
        self.vmaf_plots: bool = None
        self.probing_rate: int = None
        self.n_threads: int = None
        self.vmaf_filter: str = None

        # VVC
        self.vvc_conf: Path = None
        self.video_dimensions = (None, None)
        self.video_framerate = None

        # Set all initial values
        self.load_project(initial_data)

    def load_project(self, initial_data):
        """
        Loads project attributes to this class
        """
        # Set all initial values
        for key in initial_data:
            setattr(self, key, initial_data[key])

    def get_frames(self):
        """
        Get total frame count of input file, returning total_frames from project if already exists
        """
        # TODO: Unify get frames with vs pipe cache generation

        if self.frames > 0:
            return self.frames

        if self.chunk_method in ("vs_ffms2", "vs_lsmash"):
            vs = (
                self.input
                if self.is_vs
                else create_vs_file(self.temp, self.input, self.chunk_method)
            )
            fr = frame_probe_vspipe(vs)
            if fr > 0:
                self.frames = fr
                return fr

        total = frame_probe_fast(self.input, self.is_vs)

        self.frames = total

        return self.frames

    def set_frames(self, frame_count: int):
        """
        Setting total frame count for project
        """
        self.frames = frame_count

    def outputs_filenames(self):
        """
        Set output filename

        :param project: the Project
        """
        if self.webm:
            suffix = ".webm"
        else:
            suffix = ".mkv"

        # Check for non-empty string
        if isinstance(self.output_file, str) and self.output_file:
            if self.output_file[-1] in ('\\', '/'):
                if not Path(self.output_file).exists():
                    os.makedirs(Path(self.output_file), exist_ok=True)
                self.output_file = Path(f"{self.output_file}{self.input.stem}_{self.encoder}{suffix}")
            else:
                self.output_file = Path(self.output_file).with_suffix(suffix)
        else:
            self.output_file = Path(f"{self.input.stem}_{self.encoder}{suffix}")

    def load_project_from_file(self, path_string):
        """
        Loads projedt attributes from json to this class
        """
        pth = Path(path_string)
        with open(pth) as json_data:
            data = json.load(json_data)
        self.load_project(data)

    def save_project_to_file(self, path_string):
        """
        Save project attributes from json to this class
        """
        pth = Path(path_string)
        with open(pth, "w") as json_data:
            json_data.write(self.save_project())

    def save_project(self):
        """
        Returns json of this class, which later can be loaded
        """
        dt = dict(self.__dict__)
        del dt["input"]
        del dt["output_file"]
        del dt["temp"]
        del dt["vmaf_path"]
        del dt["config"]
        return json.dumps(dt, indent=4, sort_keys=True)

    def determine_workers(self):
        """Returns number of workers that machine can handle with selected encoder."""
        if self.workers:
            return self.workers

        cpu = os.cpu_count()
        ram = round(virtual_memory().total / 2 ** 30)

        if self.encoder in ("aom", "rav1e", "vpx"):
            workers = round(min(cpu / 3, ram / 1.5))

        elif self.encoder in ("svt_av1", "svt_vp9", "x265", "x264"):
            workers = round(min(cpu, ram)) // 8

        elif self.encoder in "vvc":
            workers = round(min(cpu, ram)) // 4

        # fix if workers round up to 0
        if workers == 0:
            workers = 1

        self.workers = workers

    def setup(self):
        """Creating temporally folders when needed."""

        if self.temp:
            if self.temp[-1] in ('\\', '/'):
                self.temp = Path(f"{self.temp}{'.' + str(hash_path(str(self.input)))}")
            else:
                self.temp = Path(str(self.temp))
        else:
            self.temp = Path("." + str(hash_path(str(self.input))))

        # Checking is resume possible
        done_path = self.temp / "done.json"
        self.resume = self.resume and done_path.exists()

        if not self.resume and self.temp.is_dir():
            shutil.rmtree(self.temp)

        (self.temp / "split").mkdir(parents=True, exist_ok=True)
        (self.temp / "encode").mkdir(exist_ok=True)

    def concat_routine(self):
        """
        Runs the concatenation routine with project

        :param project: the Project
        :return: None
        """
        try:
            if self.encoder == "vvc":
                vvc_concat(self.temp, self.output_file.with_suffix(".h266"))
            elif self.mkvmerge:
                concatenate_mkvmerge(self.temp, self.output_file)
            else:
                concatenate_ffmpeg(self.temp, self.output_file, self.encoder)
        except Exception as e:
            _, _, exc_tb = sys.exc_info()
            print(
                f"Concatenation failed, error At line: {exc_tb.tb_lineno}\nError:{str(e)}"
            )
            log(f"Concatenation failed, aborting, error: {e}")
            terminate()

    def select_best_chunking_method(self):
        """
        Selecting best chunking method based on available methods
        """
        if not find_executable("vspipe"):
            self.chunk_method = "hybrid"
            log("Set Chunking Method: Hybrid")
        else:
            try:
                import vapoursynth

                plugins = vapoursynth.get_core().get_plugins()

                if "systems.innocent.lsmas" in plugins:
                    log("Set Chunking Method: L-SMASH")
                    self.chunk_method = "vs_lsmash"

                elif "com.vapoursynth.ffms2" in plugins:
                    log("Set Chunking Method: FFMS2")
                    self.chunk_method = "vs_ffms2"

            except Exception as e:
                log(f"Vapoursynth not installed but vspipe reachable")
                log(f"Error:{e}" + "Fallback to Hybrid")
                self.chunk_method = "hybrid"

    def check_exes(self):
        """
        Checking required executables
        """

        if not find_executable("ffmpeg"):
            print("No ffmpeg")
            terminate()

        if self.chunk_method in ["vs_ffms2", "vs_lsmash"]:
            if not find_executable("vspipe"):
                print("vspipe executable not found")
                terminate()

            try:
                import vapoursynth

                plugins = vapoursynth.get_core().get_plugins()

                if (
                    self.chunk_method == "vs_lsmash"
                    and "systems.innocent.lsmas" not in plugins
                ):
                    print("lsmas is not installed")
                    terminate()

                if (
                    self.chunk_method == "vs_ffms2"
                    and "com.vapoursynth.ffms2" not in plugins
                ):
                    print("ffms2 is not installed")
                    terminate()
            except ModuleNotFoundError:
                print("Vapoursynth is not installed")
                terminate()
